from __future__ import annotations

from prettyqt.qt.QtDesigner import *  # noqa: F403

from prettyqt import qt

if qt.API == "pyqt6":
    from .pydesignercustomwidgetcollectionplugin import (
        PyDesignerCustomWidgetCollectionPlugin,
    )
    from .pydesignercustomwidgetplugin import PyDesignerCustomWidgetPlugin
else:
    from .pydesignercustomwidgetcollection import PyDesignerCustomWidgetCollection

from .abstractextensionfactory import AbstractExtensionFactory
from .pydesignertaskmenuextension import PyDesignerTaskMenuExtension
from .designercustomwidgetinterface import DesignerCustomWidgetInterface
from .designerformeditorinterface import DesignerFormEditorInterface
from .abstractformbuilder import AbstractFormBuilder
from .formbuilder import FormBuilder

__all__ = [
    "PyDesignerCustomWidgetPlugin",
    "PyDesignerCustomWidgetCollectionPlugin",
    "PyDesignerCustomWidgetCollection",
    "AbstractExtensionFactory",
    "PyDesignerTaskMenuExtension",
    "DesignerCustomWidgetInterface",
    "DesignerFormEditorInterface",
    "AbstractFormBuilder",
    "FormBuilder",
]



from __future__ import annotations

import collections
import functools
import inspect
import logging
import operator
import types
import typing


T = typing.TypeVar("T", bound=type)


logger = logging.getLogger(__name__)

# def add_docs(klass):
#     import pathlib
#     from prettyqt import paths
#     klass_name = klass.__name__
#     qclass = klass
#     for k in klass.mro():
#         if k.__name__.startswith("Q"):
#             qclass = k
#             break
#     module = k.__module__.split(".")[1]
#     path = paths.DOCSTRING_PATH /  module
#     filepath = path / f"Q{klass_name.replace('Mixin', '')}.txt"
#     if filepath.exists():
#         klass.__doc__ =  filepath.read_text()
#     return klass


def lca_type(classes: list[type]) -> type:
    return next(
        iter(
            functools.reduce(
                operator.and_,
                (collections.Counter(Klass.mro()) for Klass in classes),
            )
        )
    )


def find_common_ancestor(cls_list: list[type]) -> type:
    mros = [list(inspect.getmro(cls)) for cls in cls_list]
    track = collections.defaultdict(int)
    while mros:
        for mro in mros:
            cur = mro.pop(0)
            track[cur] += 1
            if (
                track[cur] >= len(cls_list)
                and not cur.__name__.endswith("Mixin")
                and cur.__name__.startswith("Q")
            ):
                return cur
            if len(mro) == 0:
                mros.remove(mro)
    raise TypeError("Couldnt find common base class")


def get_subclasses(klass: type, include_abstract: bool = False) -> type:
    for i in klass.__subclasses__():
        yield from get_subclasses(i)
        if include_abstract or not inspect.isabstract(i):
            yield i


def get_class_for_id(base_class: T, id_: str) -> T:
    base_classes = (
        typing.get_args(base_class)
        if isinstance(base_class, types.UnionType)
        else (base_class,)
    )
    for base_class in base_classes:
        for Klass in get_subclasses(base_class):
            if "ID" in Klass.__dict__ and id_ == Klass.ID:
                logger.debug(f"found class for id {Klass.ID!r}")
                return Klass
    raise ValueError(f"Couldnt find class with id {id_!r} for base class {base_class}")


def get_module_classes(module: types.ModuleType) -> list[type]:
    clsmembers = inspect.getmembers(module, inspect.isclass)
    return [tpl[1] for tpl in clsmembers]



from __future__ import annotations

import enum
import logging

from typing import Any

import numpy as np
import pandas as pd

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class PandasDataFrameModel(core.AbstractTableModel):
    class Roles(enum.IntEnum):
        """Role names."""

        DTypeRole = constants.USER_ROLE + 4555
        ColumnNameRole = constants.USER_ROLE + 4556

    def __init__(self, df, *args, **kwargs):
        self._multiindex_separator = " | "
        self.df = df
        super().__init__(*args, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, pd.DataFrame)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role, orientation:
            case constants.ALIGNMENT_ROLE, constants.HORIZONTAL:
                return constants.ALIGN_CENTER | constants.ALIGN_BOTTOM
            case constants.DISPLAY_ROLE, constants.HORIZONTAL:
                header = self.df.columns[section]
                if isinstance(header, tuple | list):
                    return self._multiindex_separator.join(str(i) for i in header)
                return str(header)
                # label = helpers.format_name(self.df.columns[section])
                # return f"{label}\n{self.df.iloc[:, section].dtype}"
            # case constants.DECORATION_ROLE, constants.HORIZONTAL if self.show_icons:
            #     dtype = self.df.iloc[:, section].dtype
            #     return icons.icon_for_dtype(dtype)
            case _, _:
                return None

    def set_multiindex_separator(self, separator: str):
        self._multiindex_separator = separator

    def get_multiindex_separator(self) -> str:
        return self._multiindex_separator

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        if self.df is None:
            return 0
        return 1 if type(self.df) == pd.Series else len(self.df.columns)

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        return len(self.df.index) if self.df is not None else 0

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                cell = self.df.iat[index.row(), index.column()]
                return "" if pd.isnull(cell) else cell
            case constants.TOOLTIP_ROLE:
                row = index.row()
                col = index.column()
                cell = self.df.iat[row, col]

                # NaN case
                return "NaN" if pd.isnull(cell) else str(cell)

            case self.Roles.DTypeRole:
                return self.df.iloc[:, index.column()].dtype
            case constants.USER_ROLE:
                return self.df.iloc[index.row(), index.column()]
            case self.Roles.ColumnNameRole:
                return self.df.iloc[:, index.column()].name

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        return constants.IS_EDITABLE | constants.IS_ENABLED | constants.IS_SELECTABLE

    #     return cur_flags if self.is_read_only else cur_flags | constants.IS_EDITABLE

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        match role:
            case constants.USER_ROLE | constants.EDIT_ROLE:
                row = index.row()
                col = index.column()
                try:
                    self.df.iat[row, col] = value
                except Exception as e:
                    logger.exception(e)
                    return False
                else:
                    self.dataChanged.emit(index, index)
                    return True
        return False

    def sort(self, ncol: int, order: constants.SortOrder):
        is_ascending = order == constants.ASCENDING
        with self.change_layout():
            self.df = self.df.sort_values(
                by=self.df.columns[ncol], ascending=is_ascending
            )
        self.update_all()

    multiindex_separator = core.Property(
        str, get_multiindex_separator, set_multiindex_separator
    )


class VerticalHeaderModel(core.AbstractTableModel):
    def __init__(self, df: pd.DataFrame, **kwargs):
        super().__init__(**kwargs)
        self.df = df

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        return 0 if self.df is None else self.df.index.nlevels

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        return 0 if self.df is None else self.df.index.shape[0]

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.TOOLTIP_ROLE:
                row = index.row()
                col = index.column()
                if isinstance(self.df.index, pd.MultiIndex):
                    return str(self.df.index.values[row][col])
                else:
                    return str(self.df.index.values[row])

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match role, orientation:
            case constants.DISPLAY_ROLE | constants.TOOLTIP_ROLE, constants.HORIZONTAL:
                if isinstance(self.df.index, pd.MultiIndex):
                    return str(self.df.index.names[section])
                else:
                    return str(self.df.index.name)


class HorizontalHeaderModel(core.AbstractTableModel):
    def __init__(self, df: pd.DataFrame, **kwargs):
        super().__init__(**kwargs)
        self.df = df

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        return 0 if self.df is None else self.df.columns.shape[0]

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        return 0 if self.df is None else self.df.columns.nlevels

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.TOOLTIP_ROLE:
                row = index.row()
                col = index.column()
                if isinstance(self.df.columns, pd.MultiIndex):
                    return str(self.df.columns.values[col][row])
                else:
                    return str(self.df.columns.values[col])
            case constants.ALIGNMENT_ROLE:
                return constants.ALIGN_CENTER

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role, orientation:
            case constants.DISPLAY_ROLE | constants.TOOLTIP_ROLE, constants.VERTICAL:
                if isinstance(self.df.columns, pd.MultiIndex):
                    return str(self.df.columns.names[section])
                else:
                    return str(self.df.columns.name)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    # Prepare data
    tuples = [
        ("bar", "one", "q"),
        ("bar", "two", "q"),
        ("baz", "one", "q"),
        ("baz", "two", "q"),
        ("foo", "one", "q"),
        ("foo", "two", "q"),
    ] * 10
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(60, 60), index=index, columns=index)
    table = widgets.TableView()
    table.show()
    model = PandasDataFrameModel(df, parent=table)
    model = model.proxifier[6:7, ::2].style("red", role=constants.FOREGROUND_ROLE)
    model = model.proxifier[1:4, ::2].style(
        "Courier", role=constants.FONT_ROLE, point_size=30
    )
    # model = model.proxifier[1:4, ::2].filter()
    table.set_model(model)
    print(table.get_models())
    # table.set_model(table.model().proxifier.get_proxy("color_values", parent=table))
    # model = model.proxifier.get_proxy("read_only")
    table.set_delegate("editor")

    app.exec()



from __future__ import annotations

from prettyqt import constants, core


class AbstractTableModelMixin(core.AbstractItemModelMixin):
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED
        | constants.IS_ENABLED
        | constants.IS_SELECTABLE
        | constants.NO_CHILDREN
    )

    def get_table_data(
        self,
        include_index: bool = False,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        x_range: slice | int | None = None,
        y_range: slice | int | None = None,
    ):
        match x_range:
            case None:
                colrange = range(self.columnCount())
            case slice():
                stop = x_range.stop or self.columnCount()
                colrange = range(x_range.start or 0, stop, x_range.step or 1)
            case int():
                colrange = range(x_range, x_range + 1)

        match y_range:
            case None:
                rowrange = range(self.rowCount())
            case slice():
                stop = y_range.stop or self.rowCount()
                rowrange = range(y_range.start or 0, stop, y_range.step or 1)
            case int():
                rowrange = range(y_range, y_range + 1)

        data = [[self.index(i, j).data(role) for j in colrange] for i in rowrange]
        h_header = [self.headerData(i, constants.HORIZONTAL) for i in colrange]
        v_header = (
            [self.headerData(i, constants.VERTICAL) for i in rowrange]
            if include_index
            else None
        )
        return data, h_header, v_header

    def to_dataframe(self, **kwargs):
        import pandas as pd

        data, h_header, v_header = self.get_table_data(**kwargs)
        return pd.DataFrame(data=data, columns=h_header, index=v_header)

    def to_markdown(self, use_checkstate_role: bool = True, **kwargs):
        data, h_header, v_header = self.get_table_data(**kwargs)
        if use_checkstate_role:
            kwargs["role"] = constants.CHECKSTATE_ROLE
            check_data, _, __ = self.get_table_data(**kwargs)
            for i, row in enumerate(data):
                for j, _column in enumerate(row):
                    if check_data[i][j]:
                        data[i][j] = "x"

        lines = [f"|{'|'.join(h_header)}|", f"|{'--|--'.join('' for _ in h_header)}|"]
        for row in data:
            sections = [str(i) if i else "" for i in row]
            lines.append(f"|{'|'.join(sections)}|")
        return "\n".join(lines)


class AbstractTableModel(AbstractTableModelMixin, core.QAbstractTableModel):
    pass


if __name__ == "__main__":
    from prettyqt import itemmodels, widgets

    app = widgets.app()
    model = itemmodels.QObjectPropertiesModel(app)
    table = widgets.TableView()
    table.set_model(model)
    print(model.to_markdown())



from .pandascategorylistmodel import PandasCategoryListModel
from .pandascolumnlistmodel import PandasColumnListModel, PandasIndexListModel
from .pandasindexfilterproxymodel import (
    PandasEvalFilterProxyModel,
    PandasStringColumnFilterProxyModel,
    PandasMultiStringColumnFilterProxyModel,
)
from .pandasdataframemodel import (
    PandasDataFrameModel,
    HorizontalHeaderModel,
    VerticalHeaderModel,
)


__all__ = [
    "PandasCategoryListModel",
    "PandasColumnListModel",
    "PandasIndexListModel",
    "PandasDataFrameModel",
    "VerticalHeaderModel",
    "HorizontalHeaderModel",
    "PandasEvalFilterProxyModel",
    "PandasStringColumnFilterProxyModel",
    "PandasMultiStringColumnFilterProxyModel",
]



from __future__ import annotations

import enum
import logging

from typing import Any

import polars as pl

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class PolarsDataFrameModel(core.AbstractTableModel):
    class Roles(enum.IntEnum):
        """Role names."""

        DTypeRole = constants.USER_ROLE + 4555
        ColumnNameRole = constants.USER_ROLE + 4556

    def __init__(self, df: pl.DataFrame, **kwargs):
        super().__init__(**kwargs)
        self.df = df

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, pl.DataFrame)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role, orientation:
            case constants.ALIGNMENT_ROLE, constants.HORIZONTAL:
                return constants.ALIGN_CENTER | constants.ALIGN_BOTTOM
            case constants.DISPLAY_ROLE, constants.HORIZONTAL:
                header = self.df.columns[section]
                return str(header)
            case _, _:
                return None

    def columnCount(self, parent: core.ModelIndex | None = None):
        if self.df is None:
            return 0
        return len(self.df.columns)

    def rowCount(self, parent: core.ModelIndex | None = None):
        return len(self.df) if self.df is not None else 0

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                cell = self.df[index.row(), index.column()]
                return str(cell)

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        return constants.IS_EDITABLE | constants.IS_ENABLED | constants.IS_SELECTABLE

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        match role:
            case constants.USER_ROLE | constants.EDIT_ROLE:
                row = index.row()
                col = index.column()
                try:
                    self.df.iat[row, col] = value
                except Exception as e:
                    logger.exception(e)
                    return False
                else:
                    self.dataChanged.emit(index, index)
                    return True
        return False

    def sort(self, ncol: int, order: constants.SortOrder):
        is_ascending = order == constants.ASCENDING
        with self.change_layout():
            self.df = self.df.sort_values(
                by=self.df.columns[ncol], ascending=is_ascending
            )
        self.update_all()


if __name__ == "__main__":
    from datetime import datetime

    from prettyqt import widgets

    app = widgets.app()
    df = pl.DataFrame(
        {
            "integer": [1, 2, 3],
            "date": [
                datetime(2022, 1, 1),
                datetime(2022, 1, 2),
                datetime(2022, 1, 3),
            ],
            "float": [4.0, 5.0, 6.0],
        }
    )
    table = widgets.TableView()
    table.show()
    model = PolarsDataFrameModel(df, parent=table)
    table.set_model(model)
    table.set_delegate("editor")

    app.exec()



"""Module containing custom delegate classes."""

from .buttondelegate import ButtonDelegate
from .htmlitemdelegate import HtmlItemDelegate
from .icondelegate import IconDelegate
from .nofocusdelegate import NoFocusDelegate
from .progressbardelegate import ProgressBarDelegate
from .radiodelegate import RadioDelegate
from .renderlinkdelegate import RenderLinkDelegate
from .stardelegate import StarDelegate
from .editordelegate import EditorDelegate
from .widgetdelegate import WidgetDelegate


__all__ = [
    "ButtonDelegate",
    "RadioDelegate",
    "ProgressBarDelegate",
    "IconDelegate",
    "StarDelegate",
    "RenderLinkDelegate",
    "NoFocusDelegate",
    "HtmlItemDelegate",
    "WidgetDelegate",
    "EditorDelegate",
]



from __future__ import annotations

import numpy as np
import pandas as pd

from prettyqt import constants, core, widgets
from prettyqt.qtpandas import pandasmodels, pandaswidgets


class DataFrameWidget(widgets.MainWindow):
    current_index_changed = core.Signal(core.ModelIndex)
    current_row_changed = core.Signal(int)

    def __init__(
        self,
        df: pd.DataFrame | None = None,
        object_name: str = "pandas_window",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)

        self.index_table = widgets.TableView(selection_behavior="rows")
        self.column_table = widgets.TableView(selection_behavior="rows")
        model = pandasmodels.PandasIndexListModel(df)
        self.index_table.set_model(model)
        model = pandasmodels.PandasColumnListModel(df)
        self.column_table.set_model(model)
        self.column_table.selectionModel().currentChanged.connect(self._on_current_change)

        self.data_table = pandaswidgets.DataFrameViewer(df)
        self.set_central_widget(self.data_table)

        self.category_table = widgets.TableView(selection_behavior="rows")

        self.attrs_table = widgets.TreeView(
            selection_behavior="rows", root_is_decorated=False
        )
        if df is not None:
            self.attrs_table.set_model(df.attrs)
            self.attrs_table.show_root(False)

        self.add_dockwidget(self.index_table, window_title="Index list")
        self.add_dockwidget(self.column_table, window_title="Column list")
        self.add_dockwidget(self.category_table, window_title="Category list")
        self.add_dockwidget(self.attrs_table, window_title="MetaData")

    def set_df(self, df: pd.DataFrame):
        if df is not None:
            self.attrs_table.set_model(df.attrs)
            self.attrs_table.show_root(False)
        self.data_table.set_df(df)
        model = pandasmodels.PandasIndexListModel(df)
        self.index_table.set_model(model)
        model = pandasmodels.PandasColumnListModel(df)
        self.column_table.set_model(model)
        self.category_table.set_model(None)

    def _on_current_change(self, new, old):
        role = new.model().Roles.ColumnNameRole
        column_name = new.data(role)
        df = new.data(constants.USER_ROLE)
        model = pandasmodels.PandasCategoryListModel(df, column_name)
        self.category_table.set_model(model)


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("Fusion")
    tuples = [
        ("bar", "one", "q"),
        ("bar", "two", "q"),
        ("baz", "one", "q"),
        ("baz", "two", "q"),
        ("foo", "one", "q"),
        ("foo", "two", "q"),
        ("qux", "one", "q"),
        ("qux", "two", "q"),
    ]
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(8, 8), index=index, columns=index)
    df.attrs = {"test": "test"}
    widget = DataFrameWidget(df)
    widget.show()
    app.exec()



from __future__ import annotations

import numpy as np
import pandas as pd

from prettyqt import constants, core, itemmodels, widgets
from prettyqt.qtpandas import pandasmodels


class RowsColumn(itemmodels.ColumnItem):
    name = "Rows"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return core.Locale().toString(len(item.index))
            case constants.SORT_ROLE:
                return len(item.index)


class ColumnsColumn(itemmodels.ColumnItem):
    name = "Columns"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return core.Locale().toString(len(item.columns))
            case constants.SORT_ROLE:
                return len(item.columns)


class IndexDescriptionColumn(itemmodels.ColumnItem):
    name = "Index"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                if isinstance(item.index, pd.MultiIndex):
                    return f"MultiIndex ({len(item.index.levels)})"
                return str(item.index.name)


class DataFrameListModel(itemmodels.ColumnTableModel):
    COLUMNS = [RowsColumn, ColumnsColumn, IndexDescriptionColumn]

    def __init__(self, dfs, parent=None):
        super().__init__(dfs, self.COLUMNS, parent=parent)


class DataFrameListWidget(widgets.TableView):
    def __init__(
        self,
        object_name: str = "pandas_df_manager",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        model = DataFrameListModel([])
        self.set_model(model)
        self.set_selection_behavior("rows")

    def add_df(self, df: pd.DataFrame):
        self.model().add(df)

    def _on_current_change(self, new, old):
        role = new.model().Roles.ColumnNameRole
        column_name = new.data(role)
        df = new.data(constants.USER_ROLE)
        model = pandasmodels.PandasCategoryListModel(df, column_name)
        self.category_table.set_model(model)


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("Fusion")
    i = 5
    tuples = [("bar", "one", "q"), ("bar", "two", "q")] * i
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(i * 2, i * 2), index=index, columns=index)
    df.attrs = {"test": "test"}
    widget = DataFrameListWidget()
    widget.add_df(df)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets


class GraphicsObjectMixin(core.ObjectMixin, widgets.GraphicsItemMixin):
    pass


class GraphicsObject(GraphicsObjectMixin, widgets.QGraphicsObject):
    pass



from __future__ import annotations

import pandas as pd

from prettyqt import constants, core, gui, widgets


MAX_AUTOSIZE_MS = 150  # Milliseconds given (at most) to perform column auto-sizing
MIN_TRUNC_CHARS = 8  # Minimum size (in characters) given to columns
MAX_WIDTH_CHARS = 64  # Maximum size (in characters) given to columns

SelectionFlag = core.ItemSelectionModel.SelectionFlag


class DataFrameModel:
    """Non-qt model, gets passed to our "real" models."""

    def __init__(self, data):
        super().__init__()
        self._data = data

    @property
    def shape(self) -> tuple[int, int]:
        return self._data.shape

    @property
    def header_shape(self) -> tuple[int, int]:
        return (len(self._data.columns.names), len(self._data.index.names))

    def data(self, y: int, x: int):
        return self._data.iat[y, x]

    def header(self, axis: constants.OrientationStr, x, level: int = 0):
        ax = self._data.columns if axis == "horizontal" else self._data.index
        return ax.values[x][level] if hasattr(ax, "levels") else ax.values[x]

    def name(self, axis: constants.OrientationStr, level):
        ax = self._data.columns if axis == "horizontal" else self._data.index
        return name if (name := ax.names[level]) is not None else f"L{level}"


class DataFrameDataModel(core.AbstractTableModel):
    """Model for table data."""

    def __init__(self, model, parent: widgets.QWidget | None = None):
        super().__init__(parent=parent)
        self._model = model

    def rowCount(self, index: core.ModelIndex | None = None):
        return max(1, self._model.shape[0])

    def columnCount(self, index: core.ModelIndex | None = None):
        return max(1, self._model.shape[1])

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if index.row() >= self._model.shape[0] or index.column() >= self._model.shape[1]:
            return None
        match role:
            case constants.DISPLAY_ROLE:
                val = self._model.data(index.row(), index.column())
                return "" if pd.isnull(val) else core.Locale().toString(val)
            case constants.USER_ROLE:
                return self._model.data(index.row(), index.column())


class DataFrameHeaderModel(core.AbstractTableModel):
    """Model for the two index tables."""

    def __init__(
        self,
        model,
        axis: constants.OrientationStr,
        palette: gui.QPalette,
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self._model = model
        self.axis = axis
        self._palette = palette
        if self.axis == "horizontal":
            self._shape = (self._model.header_shape[0], self._model.shape[1])
        else:
            self._shape = (self._model.shape[0], self._model.header_shape[1])

    def rowCount(self, index: core.ModelIndex | None = None):
        return max(1, self._shape[0])

    def columnCount(self, index: core.ModelIndex | None = None):
        return max(1, self._shape[1])

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role, orientation:
            case constants.ALIGNMENT_ROLE, constants.HORIZONTAL:
                return constants.ALIGN_CENTER | constants.ALIGN_BOTTOM
            case constants.ALIGNMENT_ROLE, constants.VERTICAL:
                return constants.ALIGN_RIGHT | constants.ALIGN_V_CENTER
            case constants.DISPLAY_ROLE, constants.HORIZONTAL:
                return (
                    section
                    if self.axis == "horizontal"
                    else self._model.name("vertical", section)
                )
            case constants.DISPLAY_ROLE, constants.VERTICAL:
                return (
                    section
                    if self.axis == "vertical"
                    else self._model.name("horizontal", section)
                )

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if (
            not index.isValid()
            or index.row() >= self._shape[0]
            or index.column() >= self._shape[1]
        ):
            return None
        row, col = (
            (index.row(), index.column())
            if self.axis == "horizontal"
            else (index.column(), index.row())
        )
        match role:
            case constants.BACKGROUND_ROLE:
                prev = self._model.header(self.axis, col - 1, row) if col else None
                cur = self._model.header(self.axis, col, row)
                return self._palette.midlight() if prev != cur else None
            case constants.DISPLAY_ROLE:
                val = self._model.header(self.axis, col, row)
                return "" if pd.isnull(val) else str(val)
            case constants.ALIGNMENT_ROLE:
                return constants.ALIGN_CENTER


class DataFrameLevelModel(core.AbstractTableModel):
    """Top left corner."""

    def __init__(
        self,
        model,
        palette: gui.QPalette,
        font: gui.QFont,
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self._model = model
        self._background = palette.dark().color()
        if self._background.lightness() > 127:
            self._foreground = palette.text()
        else:
            self._foreground = palette.highlightedText()
        self._palette = palette
        font.setBold(True)
        self._font = font

    def rowCount(self, index: core.ModelIndex | None = None):
        return max(1, self._model.header_shape[0])

    def columnCount(self, index: core.ModelIndex | None = None):
        return max(1, self._model.header_shape[1])

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match role, orientation:
            case constants.ALIGNMENT_ROLE, constants.HORIZONTAL:
                return constants.ALIGN_CENTER | constants.ALIGN_BOTTOM
            case constants.ALIGNMENT_ROLE, constants.VERTICAL:
                return constants.ALIGN_RIGHT | constants.ALIGN_V_CENTER
            case constants.DISPLAY_ROLE, _:
                return f"L{section}"

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        a = self._model.header_shape[0] - 1
        b = self._model.header_shape[1] - 1
        match role:
            case constants.DISPLAY_ROLE if index.row() == a and index.column() == b:
                return (
                    f"{self._model.name('vertical', index.column())} \\ "
                    f"{self._model.name('horizontal', index.column())}"
                )
            case constants.DISPLAY_ROLE if index.row() == a:
                return str(self._model.name("vertical", index.column()))
            case constants.DISPLAY_ROLE if index.column() == b:
                return str(self._model.name("horizontal", index.row()))
            case constants.FOREGROUND_ROLE:
                return self._foreground
            case constants.BACKGROUND_ROLE:
                return self._background
            case constants.FONT_ROLE:
                return self._font


class DataFrameViewer(widgets.Widget):
    def __init__(
        self, df: pd.DataFrame | None = None, parent: widgets.QWidget | None = None
    ):
        super().__init__(parent=parent)
        if df is None:
            df = pd.DataFrame()
        self._selection_rec = False
        self._model = None

        # We manually set the inactive highlight color to differentiate the
        # selection between the data/index/header. To actually make use of the
        # palette though, we also have to manually assign a new stock delegate
        # to each table view
        palette = self.get_palette()
        tmp = palette.highlight().color()
        tmp.setHsv(tmp.hsvHue(), 100, palette.midlight().color().lightness())
        palette.setBrush(
            gui.Palette.ColorGroup.Inactive, gui.Palette.ColorRole.Highlight, tmp
        )
        self.setPalette(palette)
        self.hscroll = widgets.ScrollBar("horizontal")
        self.vscroll = widgets.ScrollBar("vertical")

        self.table_level = widgets.TableView(frame_shadow="plain", show_grid=False)
        self.table_level.set_edit_triggers("none")
        self.table_level.set_scrollbar_policy("always_off")
        self.table_level.h_header.sectionResized.connect(self._index_resized)
        self.table_level.v_header.sectionResized.connect(self._header_resized)
        self.table_level.set_delegate("editor")

        self.table_header = widgets.TableView(frame_shadow="plain")
        self.table_header.v_header.hide()
        self.table_header.set_edit_triggers("none")
        self.table_header.set_scrollbar_policy("always_off")
        self.table_header.set_horizontal_scroll_mode("pixel")
        self.table_header.setHorizontalScrollBar(self.hscroll)
        self.table_header.h_header.sectionResized.connect(self._column_resized)
        self.table_header.set_delegate("editor")

        self.table_index = widgets.TableView(frame_shadow="plain")
        self.table_index.h_header.hide()
        self.table_index.set_edit_triggers("none")
        self.table_index.set_scrollbar_policy("always_off")
        self.table_index.set_vertical_scroll_mode("pixel")
        self.table_index.setVerticalScrollBar(self.vscroll)
        self.table_index.v_header.sectionResized.connect(self._row_resized)

        self.table_data = widgets.TableView(frame_shadow="plain")
        self.table_data.v_header.hide()
        self.table_data.h_header.hide()
        self.table_data.set_edit_triggers("none")
        self.table_data.set_scrollbar_policy("always_off")
        self.table_data.set_scroll_mode("pixel")
        self.table_data.setHorizontalScrollBar(self.hscroll)
        self.table_data.setVerticalScrollBar(self.vscroll)
        self.table_data.set_delegate("editor")
        layout = self.set_layout("grid", spacing=0, margin=0)
        layout[0, 0] = self.table_level
        layout[0, 1] = self.table_header
        layout[1, 0] = self.table_index
        layout[1, 1] = self.table_data
        layout[2, 1] = self.hscroll
        layout[1, 2] = self.vscroll
        # layout.addWidget(self.hscroll, 2, 0, 2, 2)
        # layout.addWidget(self.vscroll, 0, 2, 2, 2)
        self.setFocusProxy(self.table_data)

        # autosize columns on-demand
        self._autosized_cols: set[int] = set()
        self.hscroll.sliderMoved.connect(self._resize_visible_columns_to_contents)
        self.table_data.installEventFilter(self)

        avg_width = self.fontMetrics().averageCharWidth()
        self.min_trunc = avg_width * MIN_TRUNC_CHARS
        self.max_width = avg_width * MAX_WIDTH_CHARS
        if df is not None:
            self.set_df(df)

    def set_df(self, df: pd.DataFrame):
        model = DataFrameModel(df)
        self.set_model(model)

    def _select_columns(
        self,
        source: widgets.TableView,
        dest: widgets.TableView,
        deselect: widgets.TableView,
    ):
        if self._selection_rec:
            return
        self._selection_rec = True
        dsm = dest.selectionModel()
        ssm = source.selectionModel()
        dsm.clear()
        for col in (index.column() for index in ssm.selectedIndexes()):
            idx = dest.model().index(0, col)
            dsm.select(idx, SelectionFlag.Select | SelectionFlag.Columns)
        deselect.selectionModel().clear()
        self._selection_rec = False

    def _select_rows(
        self,
        source: widgets.TableView,
        dest: widgets.TableView,
        deselect: widgets.TableView,
    ):
        if self._selection_rec:
            return
        self._selection_rec = True
        dsm = dest.selectionModel()
        dsm.clear()
        indexes = source.selectionModel().selectedIndexes()
        for row in (index.row() for index in indexes):
            idx = dest.model().index(row, 0)
            dsm.select(idx, SelectionFlag.Select | SelectionFlag.Rows)
        deselect.selectionModel().clear()
        self._selection_rec = False

    def model(self):
        return self._model

    def _column_resized(self, col, _, new_width):
        self.table_data.setColumnWidth(col, new_width)
        self._update_layout()

    def _row_resized(self, row, _, new_height):
        self.table_data.setRowHeight(row, new_height)
        self._update_layout()

    def _index_resized(self, col, _, new_width):
        self.table_index.setColumnWidth(col, new_width)
        self._update_layout()

    def _header_resized(self, row, _, new_height):
        self.table_header.setRowHeight(row, new_height)
        self._update_layout()

    def _update_layout(self):
        h_width = max(
            self.table_level.v_header.sizeHint().width(),
            self.table_index.v_header.sizeHint().width(),
        )
        self.table_level.v_header.setFixedWidth(h_width)
        self.table_index.v_header.setFixedWidth(h_width)

        last_row = self._model.header_shape[0] - 1
        if last_row < 0:
            hdr_height = self.table_level.h_header.height()
        else:
            hdr_height = (
                self.table_level.rowViewportPosition(last_row)
                + self.table_level.rowHeight(last_row)
                + self.table_level.h_header.height()
            )
        self.table_header.setFixedHeight(hdr_height)
        self.table_level.setFixedHeight(hdr_height)

        last_col = self._model.header_shape[1] - 1
        if last_col < 0:
            idx_width = self.table_level.v_header.width()
        else:
            idx_width = (
                self.table_level.columnViewportPosition(last_col)
                + self.table_level.columnWidth(last_col)
                + self.table_level.v_header.width()
            )
        self.table_index.setFixedWidth(idx_width)
        self.table_level.setFixedWidth(idx_width)
        self._resize_visible_columns_to_contents()

    def set_model(self, model, relayout: bool = True):
        self._model = model
        data_model = DataFrameDataModel(model, parent=self.table_data)
        self.table_data.set_model(data_model)
        sel_model = self.table_data.selectionModel()
        sel_model.selectionChanged.connect(
            lambda *_: self._select_columns(
                self.table_data, self.table_header, self.table_level
            )
        )
        sel_model.selectionChanged.connect(
            lambda *_: self._select_rows(
                self.table_data, self.table_index, self.table_level
            )
        )
        sel_model.currentColumnChanged.connect(self._resize_current_column_to_content)

        level_model = DataFrameLevelModel(
            model, self.palette(), self.font(), parent=self.table_level
        )
        self.table_level.set_model(level_model)
        sel_model = self.table_level.selectionModel()
        sel_model.selectionChanged.connect(
            lambda *_: self._select_columns(
                self.table_level, self.table_index, self.table_data
            )
        )
        sel_model.selectionChanged.connect(
            lambda *_: self._select_rows(
                self.table_level, self.table_header, self.table_data
            )
        )

        header_model = DataFrameHeaderModel(
            model, "horizontal", self.palette(), parent=self.table_header
        )
        self.table_header.set_model(header_model)
        sel_model = self.table_header.selectionModel()
        sel_model.selectionChanged.connect(
            lambda *_: self._select_columns(
                self.table_header, self.table_data, self.table_index
            )
        )
        sel_model.selectionChanged.connect(
            lambda *_: self._select_rows(
                self.table_header, self.table_level, self.table_index
            )
        )

        index_model = DataFrameHeaderModel(model, "vertical", self.palette())
        self.table_index.set_model(index_model)
        sel_model = self.table_index.selectionModel()
        sel_model.selectionChanged.connect(
            lambda *_: self._select_rows(
                self.table_index, self.table_data, self.table_header
            )
        )
        sel_model.selectionChanged.connect(
            lambda *_: self._select_columns(
                self.table_index, self.table_level, self.table_header
            )
        )

        # needs to be called after setting all table models
        if relayout:
            self._update_layout()

    def setCurrentIndex(self, y: int, x: int):
        index = self.table_data.model().index(y, x)
        sel_model = self.table_data.selectionModel()
        sel_model.setCurrentIndex(index, SelectionFlag.ClearAndSelect)

    def _resize_column_to_contents(
        self, header: widgets.TableView, data: widgets.TableView, col: int
    ):
        hdr_width = header.get_size_hint_for_column(col)
        data_width = data.get_size_hint_for_column(col)
        if data_width > hdr_width:
            width = min(self.max_width, data_width)
        elif hdr_width > data_width * 2:
            width = max(min(hdr_width, self.min_trunc), min(self.max_width, data_width))
        else:
            width = min(self.max_width, hdr_width)
        header.setColumnWidth(col, width)

    def eventFilter(self, obj: widgets.TableView, event: core.QEvent) -> bool:
        if obj == self.table_data and event.type() == core.Event.Type.Resize:
            self._resize_visible_columns_to_contents()
        return False

    def _resize_visible_columns_to_contents(self):
        width = self._model.shape[1]
        col, end = self.table_data.get_visible_section_span("horizontal")
        while col < end:
            resized = False
            if col not in self._autosized_cols:
                self._autosized_cols.add(col)
                resized = True
                self._resize_column_to_contents(self.table_header, self.table_data, col)
            col += 1
            if resized:
                # as we resize columns, the boundary will change
                end = self.table_data.columnAt(self.rect().right())
                end = width if end == -1 else end + 1

    def _resize_current_column_to_content(self, new_index, old_index):
        if new_index.column() not in self._autosized_cols:
            # ensure the requested column is fully into view after resizing
            self._resize_visible_columns_to_contents()
            self.table_data.scrollTo(new_index)

    def resizeColumnsToContents(self):
        self._autosized_cols = set()
        to_check = min(self.table_index.model().columnCount(), 100)
        for col in range(to_check):
            self._resize_column_to_contents(self.table_level, self.table_index, col)
        self._update_layout()


if __name__ == "__main__":
    from prettyqt import debugging, eventfilters

    df = debugging.example_multiindex_df()
    app = widgets.app()
    table = DataFrameViewer()
    table.set_df(df)
    test = eventfilters.SectionAutoSpanEventFilter(table.table_header)
    test = eventfilters.SectionAutoSpanEventFilter(
        table.table_index, orientation=constants.VERTICAL
    )
    table.table_data.get_proxy("color_values")

    table.resizeColumnsToContents()
    table.show()
    app.exec()



"""svg module.

contains QtSvgWidgets-based classes
"""

from prettyqt.qt.QtSvgWidgets import *  # noqa: F403

from .graphicssvgitem import GraphicsSvgItem
from .svgwidget import SvgWidget


__all__ = ["SvgWidget", "GraphicsSvgItem"]



from __future__ import annotations

from prettyqt import constants, core


class ColumnOrderProxyModel(core.IdentityProxyModel):
    """Proxy model which reorders the columns of the source model.

    Proxy model which reorders / hides the columns of the source model by passing a list
    containing the new order. If not all indexes are part of the list, then the
    missing sections will be hidden.

    ### Example

    ```py
    table.proxifier.reorder_columns(order=[3, 2, 0])
    table.show()
    # or
    model = MyModel()
    proxy = ColumnOrderProxyModel(order=[3, 2, 0])
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "column_order"
    ICON = "mdi.reorder-vertical"

    def __init__(self, order: list[int], **kwargs):
        self._column_order = order
        super().__init__(**kwargs)

    def get_column_order(self) -> list[int]:
        return self._column_order

    def set_column_order(self, order: list[int]):
        with self.reset_model():
            self._column_order = order

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        if not proxy_index.isValid():
            return core.ModelIndex()
        return self.sourceModel().createIndex(
            proxy_index.row(),
            self._column_order[proxy_index.column()],
            proxy_index.internalPointer(),
        )

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        if not source_index.isValid():
            return core.ModelIndex()
        proxy_column = self._column_order.index(source_index.column())
        return self.createIndex(
            source_index.row(), proxy_column, source_index.internalPointer()
        )

    def sibling(self, row: int, column: int, index: core.ModelIndex) -> core.ModelIndex:
        if column >= len(self._column_order):
            return core.ModelIndex()
        return self.index(row, column, index.parent())

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole,
    ):
        if orientation == constants.HORIZONTAL:
            section = self._column_order[section]
        return self.sourceModel().headerData(section, orientation, role)

    def hasChildren(self, parent_index: core.ModelIndex) -> bool:
        if self.sourceModel() is None or parent_index.column() > 0:
            return False
        source_parent = self.mapToSource(parent_index).sibling(parent_index.row(), 0)
        return self.sourceModel().rowCount(source_parent) > 0

    def parent(self, index: core.ModelIndex) -> core.ModelIndex:
        source_parent = self.mapToSource(index).parent()
        if not source_parent.isValid():
            return core.ModelIndex()
        return self.createIndex(source_parent.row(), 0, source_parent.internalPointer())

    def index(self, row: int, column: int, parent_index: core.ModelIndex):
        if parent_index.column() > 0 or self.sourceModel() is None:
            return core.ModelIndex()
        source_parent = self.mapToSource(parent_index).sibling(parent_index.row(), 0)
        source_index = self.sourceModel().index(
            row, self._column_order[column], source_parent
        )
        if not source_index.isValid():
            return core.ModelIndex()
        return self.createIndex(row, column, source_index.internalPointer())

    def columnCount(self, index: core.ModelIndex | None = None) -> int:
        index = index or core.ModelIndex()
        return len(self._column_order)

    def rowCount(self, index: core.ModelIndex | None = None) -> int:
        index = index or core.ModelIndex()
        if self.sourceModel() is None or index.column() > 0:
            return 0
        source_parent = self.mapToSource(index).sibling(index.row(), 0)
        return self.sourceModel().rowCount(source_parent)

    order = core.Property(list, get_column_order, set_column_order)


if __name__ == "__main__":
    from prettyqt import gui, widgets

    data = dict(
        a=["abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedfaa"],
        b=[10000000, 2, 3, 4, 5, 6],
        c=[1, 2, 3, 4, 5, 6],
        d=[100000000, 2, 3, 4, 5, 6],
        e=[1000000, 2, 3, 4, 5, 6],
    )
    source_model = gui.StandardItemModel.from_dict(data)
    app = widgets.app()
    table = widgets.TableView()
    model = ColumnOrderProxyModel(order=[1, 2], parent=table)
    model.setSourceModel(source_model)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import logging

from typing import Literal

import numpy as np
import pandas as pd

from prettyqt import constants, core


logger = logging.getLogger(__name__)


FilterModeStr = Literal["startswith", "endswith", "containts", "match"]


class BasePandasIndexFilterProxyModel(core.IdentityProxyModel):
    """Base Proxy to filter a pandas dataframe based on an calculated index.

    Mappings in both directions are calculated then with numpy methods.

    Faster filter model compared to row-based SortFilterProxyModel approach.
    Doesnt need a copy of whole dataframe, only mapping indexes are built.

    There are lot of use cases though where the better option probably is to not
    use any proxy model at all, but just work on a copy of the dataframe.

    Note: Dataframe proxies cant be chained right now.
    """

    def __init__(self, **kwargs):
        self._filter_index = None
        self._source_to_proxy = None
        self._proxy_to_source = None
        self._row_count = 0
        super().__init__(**kwargs)

    def setSourceModel(self, model):
        super().setSourceModel(model)
        self._reset_filter_index(True)
        self._update_mapping()

    def _update_mapping(self):
        with self.reset_model():
            self._row_count = self._filter_index.sum()
            self._source_to_proxy = np.cumsum(self._filter_index) - 1
            self._proxy_to_source = np.where(self._filter_index == True)[0]  # noqa: E712

    def _reset_filter_index(self, init_value: bool):
        rowcount = self.sourceModel().rowCount()
        self._filter_index = np.full(rowcount, init_value, dtype=bool)

    def rowCount(self, index: core.ModelIndex | None = None) -> int:
        return self._row_count

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        """Map header data to proxy by using proxy-to-source-map."""
        if orientation == constants.HORIZONTAL:
            return super().headerData(section, orientation, role)
        return super().headerData(self._proxy_to_source[section], orientation, role)

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        source = self.sourceModel()
        if row < 0 or column < 0 or source is None or row >= self._row_count:
            return core.ModelIndex()
        source_parent = self.mapToSource(parent)
        row_pos = int(self._proxy_to_source[row])
        source_index = source.index(row_pos, column, source_parent)
        return self.mapFromSource(source_index)

    def mapToSource(self, proxy_idx: core.ModelIndex) -> core.ModelIndex:
        """Map index to source by using proxy-to-source map."""
        source = self.sourceModel()
        if (
            source is None
            or not proxy_idx.isValid()
            or proxy_idx.row() >= self._row_count
        ):
            return core.ModelIndex()
        row_pos = int(self._proxy_to_source[proxy_idx.row()])
        return source.index(row_pos, proxy_idx.column())

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        if self.sourceModel() is None or not source_index.isValid():
            return core.ModelIndex()
        row_pos = int(self._source_to_proxy[source_index.row()])
        return self.createIndex(
            row_pos, source_index.column(), source_index.internalPointer()
        )


class PandasStringColumnFilterProxyModel(BasePandasIndexFilterProxyModel):
    """Basically filters a dataframe based on df.iloc[:, column].str.somemethod(term)."""

    ID = "pandas_str_filter"

    def __init__(self, **kwargs):
        self._filter_column = 0
        self._filter_mode = "startswith"
        self._case_sensitive = True
        self._flags = 0
        self._search_term = ""
        self._na_value = False
        super().__init__(**kwargs)

    def set_search_term(self, search_term: str):
        self._search_term = search_term
        if not self._search_term:
            self._reset_filter_index(init_value=True)
            self._update_mapping()
            return
        df = self.get_source_model(skip_proxies=True).df
        match self.filter_mode:
            case "startswith":
                self._filter_index = df.iloc[:, self._filter_column].str.startswith(
                    self._search_term, na=self._na_value
                )
            case "endswith":
                self._filter_index = df.iloc[:, self._filter_column].str.endswith(
                    self._search_term, na=self._na_value
                )
            case "contains":
                self._filter_index = df.iloc[:, self._filter_column].str.contains(
                    self._search_term,
                    case=self._case_sensitive,
                    flags=self._flags,
                    na=self._na_value,
                )
            case "match":
                self._filter_index = df.iloc[:, self._filter_column].str.match(
                    self._search_term,
                    case=self._case_sensitive,
                    flags=self._flags,
                    na=self._na_value,
                )
        self._filter_index = self._filter_index.to_numpy()
        # this is needed for new StringDtype, otherwise much slower.
        if self._filter_index.dtype == object:
            self._filter_index = self._filter_index.astype(bool)
        self._update_mapping()

    def get_search_term(self) -> str:
        return self._search_term

    def set_filter_column(self, column: int):
        self._filter_column = column

    def get_filter_column(self) -> int:
        return self._filter_column

    def set_filter_mode(self, mode: FilterModeStr):
        self._filter_mode = mode

    def get_filter_mode(self) -> FilterModeStr:
        return self._filter_mode

    def set_case_sensitive(self, mode: bool):
        self._case_sensitive = mode

    def is_case_sensitive(self) -> bool:
        return self._case_sensitive

    def set_flags(self, flags: int):
        self._flags = flags

    def get_flags(self) -> int:
        return self._flags

    def set_na_value(self, value: bool):
        self._na_value = value

    def get_na_value(self) -> bool:
        return self._na_value

    search_term = core.Property(str, get_search_term, set_search_term)
    filter_column = core.Property(int, get_filter_column, set_filter_column)
    filter_mode = core.Property(str, get_filter_mode, set_filter_mode)
    case_sensitive = core.Property(bool, is_case_sensitive, set_case_sensitive)
    re_flags = core.Property(int, get_flags, set_flags)
    na_value = core.Property(bool, get_na_value, set_na_value)


class PandasEvalFilterProxyModel(BasePandasIndexFilterProxyModel):
    ID = "pandas_eval_filter"

    def __init__(self, **kwargs):
        self._expression = ""
        super().__init__(**kwargs)

    def set_expression(self, expression: str):
        self._expression = expression
        if not self._expression:
            self._reset_filter_index(True)
            self._update_mapping()
        try:
            self._filter_index = df.eval(self._expression)
            self._filter_index = self._filter_index.to_numpy()
            # df.eval doesnt neccessarily return a bool index. If not, show nothing.
            if self._filter_index.dtype != bool:
                self._reset_filter_index(False)
        except Exception:
            self._reset_filter_index(False)
        self._update_mapping()

    def get_expression(self) -> str:
        return self._expression

    expression = core.Property(str, get_expression, set_expression)


class PandasMultiStringColumnFilterProxyModel(BasePandasIndexFilterProxyModel):
    def __init__(self, **kwargs):
        self._filters: dict[str, str] = {}
        super().__init__(**kwargs)

    def set_filters(self, filters: dict[str, str]):
        self._filters = filters
        df = self.get_source_model(skip_proxies=True).df
        # workaround-ish way to implement "startswith" as an expression
        filters = [f"('{v}' <= `{k}` <= '{v}~')" for k, v in self._filters.items()]
        expr = " & ".join(filters)
        try:
            self._filter_index = df.eval(expr)
            self._filter_index = self._filter_index.to_numpy()
            if self._filter_index.dtype != bool:
                self._filter_index = self._filter_index.astype(bool)
        except Exception:
            self._reset_filter_index(False)
        self._update_mapping()

    def get_filters(self) -> dict[str, str]:
        return self._filters

    filters = core.Property(dict, get_filters, set_filters)


if __name__ == "__main__":
    from prettyqt import debugging, widgets
    from prettyqt.qtpandas import pandasmodels

    app = widgets.app()
    a = pd.Series(["a", "bc", "c", "d", "aa"] * 100000, dtype=pd.StringDtype())
    b = pd.Series(["a", "b", "c", "fjkdsj", "fdf"] * 100000, dtype=pd.StringDtype())
    df = pd.DataFrame(dict(a=a, b=b))
    model = pandasmodels.PandasDataFrameModel(df)
    proxy = PandasStringColumnFilterProxyModel(parent=model)
    proxy.setSourceModel(model)
    w = debugging.ProxyComparerWidget(proxy)
    w.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Sequence
import logging

from typing_extensions import Self

from prettyqt import core
from prettyqt.utils import get_repr


logger = logging.getLogger(__name__)


class TreeItem:
    """Tree node class that can be used to build trees of objects."""

    __slots__ = ("parent_item", "obj", "children", "has_children", "children_fetched")

    def __init__(self, obj, parent: Self | None = None):
        self.parent_item = parent
        self.obj = obj
        self.children: list[Self] = []
        self.has_children = True
        self.children_fetched = False

    def __repr__(self):
        return get_repr(self, self.obj)

    def __iter__(self) -> Iterator[Self]:
        return iter(self.children)

    def __getitem__(self, index: int) -> Self:
        return self.children[index]

    def __rshift__(self, other: Self):
        """Set children using >> bitshift operator for self >> other.

        Args:
            other (Self): other node, children
        """
        other.parent_item = self

    def __lshift__(self, other: Self):
        """Set parent using << bitshift operator for self << other.

        Args:
            other (Self): other node, parent
        """
        self.parent_item = other

    def __copy__(self) -> Self:
        """Shallow copy self."""
        obj = type(self).__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        return obj

    def append_child(self, item: Self):
        item.parent_item = self
        self.children.append(item)

    def insert_children(self, idx: int, items: Sequence[Self]):
        self.children[idx:idx] = items
        for item in items:
            item.parent_item = self

    def parent(self) -> Self | None:
        return self.parent_item

    @property
    def ancestors(self) -> Iterable[Self]:
        """Get iterator to yield all ancestors of self, does not include self."""
        node = self
        while (node := node.parent_item) is not None:
            yield node

    @property
    def descendants(self) -> Iterable[Self]:
        """Get iterator to yield all descendants of self, does not include self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node != self)

    @property
    def leaves(self) -> Iterable[Self]:
        """Get iterator to yield all leaf nodes from self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node.is_leaf)

    @property
    def siblings(self) -> Iterable[Self]:
        """Get siblings of self."""
        if self.parent_item is None:
            return ()
        return tuple(child for child in self.parent_item.children if child is not self)

    @property
    def left_sibling(self) -> Self | None:
        """Get sibling left of self."""
        if self.parent_item:
            children = self.parent_item.children
            if child_idx := children.index(self):
                return self.parent_item.children[child_idx - 1]

    @property
    def right_sibling(self) -> Self | None:
        """Get sibling right of self."""
        if self.parent_item:
            children = self.parent_item.children
            child_idx = children.index(self)
            if child_idx + 1 < len(children):
                return self.parent_item.children[child_idx + 1]

    @property
    def node_path(self) -> Iterable[Self]:
        """Get tuple of nodes starting from root."""
        if self.parent_item is None:
            return [self]
        return (*list(self.parent_item.node_path), self)

    @property
    def is_root(self) -> bool:
        """Get indicator if self is root node."""
        return self.parent_item is None

    @property
    def is_leaf(self) -> bool:
        """Get indicator if self is leaf node."""
        return not len(list(self.children))

    @property
    def root(self) -> Self:
        """Get root node of tree."""
        return self if self.parent_item is None else self.parent_item.root

    @property
    def depth(self) -> int:
        """Get depth of self, indexing starts from 1."""
        return 1 if self.parent_item is None else self.parent_item.depth + 1

    @property
    def max_depth(self) -> int:
        """Get maximum depth from root to leaf node."""
        return max(
            [self.root.depth] + [node.depth for node in list(self.root.descendants)]
        )

    def row(self) -> int:
        return self.parent_item.children.index(self) if self.parent_item else 0

    def pretty_print(self, indent: int = 0):
        text = indent * "    " + str(self)
        logger.debug(text)
        for child_item in self.children:
            child_item.pretty_print(indent + 1)


class TreeModel(core.AbstractItemModel):
    TreeItem = TreeItem

    def __init__(self, obj=None, show_root: bool = True, **kwargs):
        super().__init__(**kwargs)
        self._root_item = self.TreeItem(obj=obj)
        self._show_root = show_root
        self.set_root_item(obj)

    @property
    def show_root(self) -> bool:
        """Return True if the inspected node is visible.

        In that case an invisible root node has been added.
        """
        return self._show_root

    def set_root_item(self, obj):
        if self._show_root:
            self._root_item = self.TreeItem(obj=None)
            self._root_item.children_fetched = True
            self.inspected_item = self.TreeItem(obj=obj)
            self._root_item.append_child(self.inspected_item)
            # root_index = self.index(0, 0)
            # self.fetchMore(self.index(0, 0, root_index))
        else:
            # The root itself will be invisible
            self._root_item = self.TreeItem(obj=obj)
            self.inspected_item = self._root_item
            root_index = self.index(0, 0)
            self.fetchMore(root_index)

    @property
    def root_item(self) -> TreeModel.TreeItem:
        return self._root_item

    def data_by_index(self, index: core.ModelIndex) -> TreeModel.TreeItem:
        return index.internalPointer() if index.isValid() else self.root_item

    def index_for_item(
        self, item: TreeModel.TreeItem, column: int = 0
    ) -> core.ModelIndex:
        if item == self.root_item:
            return core.ModelIndex()
        parent = item.parent_item or self.root_item
        if (not parent) or (len(parent.children) == 0):
            return None
        row = parent.children.index(item)
        return self.createIndex(row, column, item)

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        if not self.hasIndex(row, column, parent):
            return core.ModelIndex()
        parent_item = self.data_by_index(parent)
        child_item = parent_item.children[row]
        return self.createIndex(row, column, child_item)

    def parent(self, index: core.ModelIndex | None = None) -> core.ModelIndex:
        # hacky way to let the case without any arguments get through.
        # not really nice, a proper dispatch library would be better.
        # functools.singledispatchmethod doesnt work here.
        index = index or core.ModelIndex()
        if not index.isValid():
            return core.ModelIndex()

        child_item = self.data_by_index(index)
        parent_item = child_item.parent()

        if parent_item is None or parent_item == self.root_item:
            return core.ModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if parent.column() > 0 else len(self.data_by_index(parent).children)

    def hasChildren(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        treeitem = self.data_by_index(parent)
        if self._show_root and treeitem == self._root_item:
            return True
        return self._has_children(treeitem)

    def canFetchMore(self, parent: core.ModelIndex | None = None):
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        else:
            return not self.data_by_index(parent).children_fetched

    def fetchMore(self, parent: core.ModelIndex | None = None):
        """Fetch the children given the model index of a parent node.

        Adds the children to the parent.
        """
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return

        parent_item = self.data_by_index(parent)
        if parent_item.children_fetched:
            return

        tree_items = self._fetch_object_children(parent_item)

        with self.insert_rows(0, len(tree_items) - 1, parent):
            for tree_item in tree_items:
                parent_item.append_child(tree_item)
            parent_item.children_fetched = True

    def _fetch_object_children(self, treeitem) -> list[TreeModel.TreeItem]:
        return treeitem.children

    def _has_children(self, treeitem) -> bool:
        return treeitem.has_children


def preorder_iter(
    tree: TreeItem,
    filter_condition: Callable[[TreeItem], bool] | None = None,
    stop_condition: Callable[[TreeItem], bool] | None = None,
    max_depth: int = 0,
) -> Iterable[TreeItem]:
    """Iterate through all children of a tree.

    Pre-Order Iteration Algorithm, NLR
        1. Visit the current node.
        2. Recursively traverse the current node's left subtree.
        3. Recursively traverse the current node's right subtree.

    It is topologically sorted because a parent node is processed before its child nodes.

    >>> path_list = ["a/b/d", "a/b/e/g", "a/b/e/h", "a/c/f"]
    >>> root = list_to_tree(path_list)
    >>> print_tree(root)
    a
    ├── b
    │   ├── d
    │   └── e
    │       ├── g
    │       └── h
    └── c
        └── f

    >>> [node.node_name for node in preorder_iter(root)]
    ['a', 'b', 'd', 'e', 'g', 'h', 'c', 'f']

    >>> [node.node_name for node in preorder_iter(root,
    filter_condition=lambda x: x.node_name in ["a", "d", "e", "f", "g"])]
    ['a', 'd', 'e', 'g', 'f']

    >>> [node.node_name for node in preorder_iter(root,
    stop_condition=lambda x: x.node_name=="e")]
    ['a', 'b', 'd', 'c', 'f']

    >>> [node.node_name for node in preorder_iter(root, max_depth=3)]
    ['a', 'b', 'd', 'e', 'c', 'f']

    Args:
        tree: input tree
        filter_condition: function that takes in node as argument, optional
            Return node if condition evaluates to `True`
        stop_condition: function that takes in node as argument, optional
            Stops iteration if condition evaluates to `True`
        max_depth: maximum depth of iteration, based on `depth` attribute, optional
    """
    if (
        tree
        and (not max_depth or tree.depth <= max_depth)
        and (not stop_condition or not stop_condition(tree))
    ):
        if not filter_condition or filter_condition(tree):
            yield tree
        for child in tree.children:
            yield from preorder_iter(child, filter_condition, stop_condition, max_depth)


if __name__ == "__main__":
    model = TreeModel("test")
    model.parent()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


SortIndicator = widgets.QStyleOptionHeader.SortIndicator
StateFlag = widgets.QStyle.StateFlag
CE = widgets.QStyle.ControlElement
SelectedPosition = widgets.QStyleOptionHeader.SelectedPosition
SectionPosition = widgets.QStyleOptionHeader.SectionPosition

HORIZONTAL_HEADER_DATA_ROLE = constants.ItemDataRole.UserRole + 150
VERTICAL_HEADER_DATA_ROLE = constants.ItemDataRole.UserRole + 151


def find_root_index(index: core.ModelIndex) -> core.ModelIndex:
    while index.parent().isValid():
        index = index.parent()
    return index


def get_parent_indexes(index: core.ModelIndex) -> list[core.ModelIndex]:
    indexes = []
    while index.isValid():
        indexes.insert(0, index)
        index = index.parent()
    return indexes


class HierarchicalHeaderView(widgets.HeaderView):
    """Hierarchical header view.

    This class is a Python port of
    http://qt-apps.org/content/show.php/HierarchicalHeaderView?content=103154
    """

    header_model = None

    def __init__(
        self,
        orientation: constants.Orientation | constants.OrientationStr,
        parent: widgets.QWidget,
    ):
        super().__init__(orientation, parent, highlight_sections=True)
        self.setSectionsClickable(True)
        self.sectionResized.connect(self.on_section_resized)
        self.show()  # force to be visible
        if orientation in {constants.HORIZONTAL, "horizontal"}:
            parent.setHorizontalHeader(self)
        else:
            parent.setVerticalHeader(self)
        self.sectionMoved.connect(self._on_section_moved)

    def init_from_new_model(
        self, orientation: constants.Orientation, model: core.QAbstractItemModel
    ):
        is_hor = orientation == constants.HORIZONTAL
        role = HORIZONTAL_HEADER_DATA_ROLE if is_hor else VERTICAL_HEADER_DATA_ROLE
        self.header_model = model.data(core.ModelIndex(), role)

    def find_leaf(
        self, index: core.ModelIndex, section_index: int, current_leaf_index: int
    ) -> tuple[core.ModelIndex, int]:
        if not index.isValid():
            return core.ModelIndex(), current_leaf_index
        model = index.model()
        if child_count := model.columnCount(index):
            for i in range(child_count):
                res, current_leaf_index = self.find_leaf(
                    model.index(0, i, index), section_index, current_leaf_index
                )
                if res.isValid():
                    return res, current_leaf_index
        else:
            current_leaf_index += 1
            if current_leaf_index == section_index:
                return index, current_leaf_index
        return core.ModelIndex(), current_leaf_index

    def leaf_index(self, section_index: int) -> core.ModelIndex:
        if self.header_model:
            current_leaf_index = -1
            for i in range(self.header_model.columnCount()):
                res, current_leaf_index = self.find_leaf(
                    self.header_model.index(0, i), section_index, current_leaf_index
                )
                if res.isValid():
                    return res
        return core.ModelIndex()

    def search_leafs(self, index: core.ModelIndex) -> list[core.ModelIndex]:
        res = []
        if index.isValid():
            model = index.model()
            if child_count := index.model().columnCount(index):
                for i in range(child_count):
                    leaf = model.index(0, i, index)
                    res += self.search_leafs(leaf)
            else:
                res.append(index)
        return res

    def leafs(self, index: core.ModelIndex) -> list[core.ModelIndex]:
        leafs = []
        if index.isValid():
            model = index.model()
            child_count = index.model().columnCount(index)
            for i in range(child_count):
                leaf = model.index(0, i, index)
                leafs += self.search_leafs(leaf)
        return leafs

    def set_foreground_brush(
        self, opt: widgets.QStyleOptionHeader, index: core.ModelIndex
    ):
        if foreground_brush := index.data(constants.FOREGROUND_ROLE):
            brush = gui.QBrush(foreground_brush)
            opt.palette.setBrush(gui.QPalette.ColorRole.ButtonText, brush)

    def set_background_brush(
        self, opt: widgets.QStyleOptionHeader, index: core.ModelIndex
    ):
        if background_brush := index.data(constants.BACKGROUND_ROLE):
            brush = gui.QBrush(background_brush)
            opt.palette.setBrush(gui.QPalette.ColorRole.Button, brush)
            opt.palette.setBrush(gui.QPalette.ColorRole.Window, brush)

    def get_cell_size(
        self,
        leaf_index: core.ModelIndex,
        hv: widgets.QHeaderView,
        style_options: widgets.QStyleOptionHeader,
    ) -> core.QSize:
        res = core.QSize()
        if variant := leaf_index.data(constants.SIZE_HINT_ROLE):
            res = variant
        fnt = var if (var := leaf_index.data(constants.FONT_ROLE)) else hv.font()
        fnt.setBold(True)
        fm = gui.QFontMetrics(fnt)
        text_size = fm.size(0, leaf_index.data(constants.DISPLAY_ROLE))
        size = text_size + core.QSize(4, 0)
        if leaf_index.data(constants.USER_ROLE):
            size.transpose()
        decoration_size = hv.style().sizeFromContents(
            widgets.QStyle.ContentsType.CT_HeaderSection,
            style_options,
            core.QSize(),
            hv,
        )
        empty_text_size = fm.size(0, "")
        return res.expandedTo(size + decoration_size - empty_text_size)

    def get_current_cell_width(
        self,
        searched_index: core.ModelIndex,
        leaf_index: core.ModelIndex,
        section_index: int,
        hv: widgets.QHeaderView,
    ) -> int:
        leafs_list = self.leafs(searched_index)
        if not leafs_list:
            return hv.sectionSize(section_index)
        offset = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
        first_leaf_section_index = section_index - offset
        return sum(
            hv.sectionSize(first_leaf_section_index + i) for i in range(len(leafs_list))
        )

    def get_current_cell_left(
        self,
        searched_index: core.ModelIndex,
        leaf_index: core.ModelIndex,
        section_index: int,
        left: int,
        hv: widgets.QHeaderView,
    ) -> int:
        if leafs_list := self.leafs(searched_index):
            n = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
            first_leaf_section_index = section_index - n
            for i in range(n - 1, -1, -1):
                left -= hv.sectionSize(first_leaf_section_index + i)
        return left

    def paint_horizontal_cell(
        self,
        painter: gui.QPainter,
        hv: widgets.QHeaderView,
        cell_index: core.ModelIndex,
        leaf_index: core.ModelIndex,
        logical_leaf_index: int,
        style_options: widgets.QStyleOptionHeader,
        section_rect: core.QRect,
        top: int,
    ) -> int:
        uniopt = widgets.QStyleOptionHeader(style_options)
        self.set_foreground_brush(uniopt, cell_index)
        self.set_background_brush(uniopt, cell_index)
        height = (
            section_rect.height() - top
            if cell_index == leaf_index
            else self.get_cell_size(cell_index, hv, uniopt).height()
        )
        left = self.get_current_cell_left(
            cell_index, leaf_index, logical_leaf_index, section_rect.left(), hv
        )
        width = self.get_current_cell_width(
            cell_index, leaf_index, logical_leaf_index, hv
        )
        r = core.QRect(left, top, width, height)
        uniopt.text = cell_index.data(constants.DISPLAY_ROLE)
        painter.save()
        uniopt.rect = r
        style = hv.style()
        if cell_index.data(constants.USER_ROLE):
            style.drawControl(CE.CE_HeaderSection, uniopt, painter, hv)
            m = gui.QTransform()
            m.rotate(-90)
            painter.setWorldTransform(m, True)
            new_r = core.QRect(0, 0, r.height(), r.width())
            new_r.moveCenter(core.QPoint(-r.center().y(), r.center().x()))
            uniopt.rect = new_r
            style.drawControl(CE.CE_HeaderLabel, uniopt, painter, hv)
        else:
            style.drawControl(CE.CE_Header, uniopt, painter, hv)
        painter.restore()
        return top + height

    def paint_horizontal_section(
        self,
        painter: gui.QPainter,
        section_rect: core.QRect,
        logical_leaf_index: int,
        hv: widgets.QHeaderView,
        style_options: widgets.QStyleOptionHeader,
        leaf_index: core.ModelIndex,
    ):
        old_bo = painter.brushOrigin()
        top = section_rect.y()
        indexes = get_parent_indexes(leaf_index)
        for i, idx in enumerate(indexes):
            real_style_options = widgets.QStyleOptionHeader(style_options)
            if i < len(indexes) - 1 and (
                real_style_options.state & StateFlag.State_Sunken
                or real_style_options.state & StateFlag.State_On
            ):
                t = StateFlag.State_Sunken | StateFlag.State_On
                real_style_options.state = real_style_options.state & ~t
                # FIXME: parent items are not highlighted
            if i < len(indexes) - 1:  # Use sortIndicator for inner level only
                real_style_options.sortIndicator = SortIndicator.None_
            #                if i==0:
            #                    print(self.leafs(indexes[i]), leaf_index)
            top = self.paint_horizontal_cell(
                painter,
                hv,
                idx,
                leaf_index,
                logical_leaf_index,
                real_style_options,
                section_rect,
                top,
            )
        painter.setBrushOrigin(old_bo)

    def paint_vertical_cell(
        self,
        painter: gui.QPainter,
        hv: widgets.QHeaderView,
        cell_index: core.ModelIndex,
        leaf_index: core.ModelIndex,
        logical_leaf_index: int,
        style_options: widgets.QStyleOptionHeader,
        section_rect: core.QRect,
        left: int,
    ) -> int:
        uniopt = widgets.QStyleOptionHeader(style_options)
        self.set_foreground_brush(uniopt, cell_index)
        self.set_background_brush(uniopt, cell_index)
        width = (
            section_rect.width() - left
            if cell_index == leaf_index
            else self.get_cell_size(cell_index, hv, uniopt).width()
        )
        top = self.get_current_cell_left(
            cell_index, leaf_index, logical_leaf_index, section_rect.top(), hv
        )
        height = self.get_current_cell_width(
            cell_index, leaf_index, logical_leaf_index, hv
        )
        r = core.QRect(left, top, width, height)
        uniopt.text = cell_index.data(constants.DISPLAY_ROLE)
        painter.save()
        uniopt.rect = r
        style = hv.style()
        if cell_index.data(constants.USER_ROLE):
            style.drawControl(CE.CE_HeaderSection, uniopt, painter, hv)
            m = gui.QTransform()
            m.rotate(-90)
            painter.setWorldTransform(m, True)
            new_r = core.QRect(0, 0, r.height(), r.width())
            new_r.moveCenter(core.QPoint(-r.center().y(), r.center().x()))
            uniopt.rect = new_r
            style.drawControl(CE.CE_HeaderLabel, uniopt, painter, hv)
        else:
            style.drawControl(CE.CE_Header, uniopt, painter, hv)
        painter.restore()
        return left + width

    def paint_vertical_section(
        self,
        painter: gui.QPainter,
        section_rect: core.QRect,
        logical_leaf_index: int,
        hv: widgets.QHeaderView,
        style_options: widgets.QStyleOptionHeader,
        leaf_index: core.ModelIndex,
    ):
        old_bo = painter.brushOrigin()
        left = section_rect.x()
        indexes = get_parent_indexes(leaf_index)
        for i, idx in enumerate(indexes):
            real_style_options = widgets.QStyleOptionHeader(style_options)
            if i < len(indexes) - 1 and (
                real_style_options.state & StateFlag.State_Sunken
                or real_style_options.state & StateFlag.State_On
            ):
                t = StateFlag.State_Sunken | StateFlag.State_On
                real_style_options.state = real_style_options.state & ~t
                # FIXME: parent items are not highlighted
            left = self.paint_vertical_cell(
                painter,
                hv,
                idx,
                leaf_index,
                logical_leaf_index,
                real_style_options,
                section_rect,
                left,
            )
        painter.setBrushOrigin(old_bo)

    def _on_section_moved(
        self, logical_index: int, old_visual_index: int, new_visual_index: int
    ):
        view = self.parent()
        model = view.model()
        if not hasattr(model, "reorder"):
            return  # reorder underlying data of models with /reorder/ def only
        if getattr(self, "manual_move", False):
            self.manual_move = False
            return
        self.manual_move = True
        self.moveSection(new_visual_index, old_visual_index)  # cancel move
        if not model.reorder(old_visual_index, new_visual_index, self.orientation()):
            return
        rng = sorted((old_visual_index, new_visual_index))

        if self.orientation() == constants.HORIZONTAL:
            options = [(view.columnWidth(i), i) for i in range(rng[0], rng[1] + 1)]
            for i, col in enumerate(range(rng[0], rng[1] + 1)):
                view.setColumnWidth(col, options[i][0])
            view.selectColumn(new_visual_index)
        else:
            options = [(view.rowHeight(i), i) for i in range(rng[0], rng[1] + 1)]
            for i, col in enumerate(range(rng[0], rng[1] + 1)):
                view.setRowHeight(col, options[i][0])
            view.selectRow(new_visual_index)

        # FIXME: don't select if sorting is enable?
        if self.isSortIndicatorShown():
            sort_ind_index = next(
                (i for i, o in enumerate(options) if o[1] == self.sortIndicatorSection()),
                None,
            )
            # sort indicator is among sections being reordered
            if sort_ind_index is not None:
                # FIXME: does unnecessary sorting
                self.setSortIndicator(sort_ind_index + rng[0], self.sortIndicatorOrder())
        model.layoutChanged.emit()  # update view

    def get_style_option_for_cell(self, logical_index: int) -> widgets.QStyleOptionHeader:
        opt = widgets.QStyleOptionHeader()
        self.initStyleOption(opt)
        if self.isSortIndicatorShown() and self.sortIndicatorSection() == logical_index:
            asc = self.sortIndicatorOrder() == constants.ASCENDING
            opt.sortIndicator = SortIndicator.SortDown if asc else SortIndicator.SortUp
        if self.window().isActiveWindow():
            opt.state = opt.state | StateFlag.State_Active
        opt.textAlignment = constants.AlignmentFlag.AlignCenter
        opt.iconAlignment = constants.AlignmentFlag.AlignVCenter
        opt.section = logical_index
        visual = self.visualIndex(logical_index)
        if self.count() == 1:
            opt.position = SectionPosition.OnlyOneSection
        elif visual == 0:
            opt.position = SectionPosition.Beginning
        else:
            is_end = visual == self.count() - 1
            opt.position = SectionPosition.End if is_end else SectionPosition.Middle
        sel_model = self.selectionModel()
        if not sel_model:
            return opt
        root_idx = self.rootIndex()
        if self.sectionsClickable() and self.highlightSections():
            if self.orientation() == constants.HORIZONTAL:
                if sel_model.columnIntersectsSelection(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_On
                if sel_model.isColumnSelected(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_Sunken
            else:
                if sel_model.rowIntersectsSelection(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_On
                if sel_model.isRowSelected(logical_index, root_idx):
                    opt.state = opt.state | StateFlag.State_Sunken
        prev_idx = self.logicalIndex(visual - 1)
        next_idx = self.logicalIndex(visual + 1)
        if self.orientation() == constants.HORIZONTAL:
            prev_selected = sel_model.isColumnSelected(prev_idx, root_idx)
            next_selected = sel_model.isColumnSelected(next_idx, root_idx)
        else:
            prev_selected = sel_model.isRowSelected(prev_idx, root_idx)
            next_selected = sel_model.isRowSelected(next_idx, root_idx)

        if prev_selected and next_selected:
            opt.selectedPosition = SelectedPosition.NextAndPreviousAreSelected
        elif prev_selected:
            opt.selectedPosition = SelectedPosition.PreviousIsSelected
        elif next_selected:
            opt.selectedPosition = SelectedPosition.NextIsSelected
        else:
            opt.selectedPosition = SelectedPosition.NotAdjacent
        return opt

    def sectionSizeFromContents(self, logical_index: int) -> core.QSize:
        if not self.header_model:
            return super().sectionSizeFromContents(logical_index)
        cur_leaf_index = self.leaf_index(logical_index)
        if not cur_leaf_index.isValid():
            return super().sectionSizeFromContents(logical_index)
        styleOption = widgets.QStyleOptionHeader(
            self.get_style_option_for_cell(logical_index)
        )
        s = self.get_cell_size(cur_leaf_index, self, styleOption)
        cur_leaf_index = cur_leaf_index.parent()
        while cur_leaf_index.isValid():
            cell_size = self.get_cell_size(cur_leaf_index, self, styleOption)
            if self.orientation() == constants.HORIZONTAL:
                s.setHeight(s.height() + cell_size.height())
            else:
                s.setWidth(s.width() + cell_size.width())
            cur_leaf_index = cur_leaf_index.parent()
        return s

    def paintSection(self, painter: gui.QPainter, rect: core.QRect, logical_index: int):
        if not rect.isValid():
            super().paintSection(painter, rect, logical_index)
            return
        leaf_index = self.leaf_index(logical_index)
        if not leaf_index.isValid():
            super().paintSection(painter, rect, logical_index)
            return
        style_option = self.get_style_option_for_cell(logical_index)
        if self.orientation() == constants.HORIZONTAL:
            self.paint_horizontal_section(
                painter,
                rect,
                logical_index,
                self,
                style_option,
                leaf_index,
            )
        else:
            self.paint_vertical_section(
                painter,
                rect,
                logical_index,
                self,
                style_option,
                leaf_index,
            )

    def on_section_resized(self, logical_index: int):
        if self.isSectionHidden(logical_index):
            return
        leaf_index = self.leaf_index(logical_index)
        if leaf_index.isValid():
            leafs_list = self.leafs(find_root_index(leaf_index))
            start = leafs_list.index(leaf_index) if leaf_index in leafs_list else -1
            is_horizontal = self.orientation() == constants.HORIZONTAL
            for _ in range(start, 0, -1):
                logical_index -= 1
                w = self.viewport().width()
                h = self.viewport().height()
                pos = self.sectionViewportPosition(logical_index)
                r = core.QRect(pos, 0, w - pos, h)
                if is_horizontal:
                    if self.isRightToLeft():
                        r.setRect(0, 0, pos + self.sectionSize(logical_index), h)
                else:
                    r.setRect(0, pos, w, h - pos)
                self.viewport().update(r.normalized())

    def setModel(self, model: core.QAbstractItemModel):
        super().setModel(model)
        model.layoutChanged.connect(self._on_layout_change)
        self._on_layout_change()

    def _on_layout_change(self):
        if model := self.model():
            self.init_from_new_model(self.orientation(), model)
            is_horizontal = self.orientation() == constants.HORIZONTAL
            count = model.columnCount() if is_horizontal else model.rowCount()
            self.initializeSections(0, count - 1)


if __name__ == "__main__":
    import numpy as np
    import pandas as pd

    from prettyqt.qtpandas import pandasmodels

    class DataFrameModel(pandasmodels.PandasDataFrameModel):
        def read_level(
            self, y: int = 0, xs: int = 0, xe: int | None = None, orient=None
        ) -> list[gui.StandardItem]:
            c = (
                self.df.columns
                if orient == HORIZONTAL_HEADER_DATA_ROLE
                else self.df.index
            )
            if not hasattr(c, "levels"):  # not MultiIndex
                return [gui.StandardItem(str(i)) for i in c]
            sibl = []
            section_start, v, xe = xs, None, xe or len(c)
            for i in range(xs, xe):
                label = c.codes[y][i]
                if label != v:
                    if y + 1 < len(c.levels) and i > xs:
                        children = self.read_level(y + 1, section_start, i, orient=orient)
                        sibl[-1].appendRow(children)
                    item = gui.StandardItem(str(c.levels[y][label]))
                    sibl.append(item)
                    section_start = i
                    v = label
            if y + 1 < len(c.levels):
                children = self.read_level(y + 1, section_start, orient=orient)
                sibl[-1].appendRow(children)
            return sibl

        def data(
            self,
            index: core.ModelIndex,
            role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        ):
            if role in (HORIZONTAL_HEADER_DATA_ROLE, VERTICAL_HEADER_DATA_ROLE):
                hm = gui.StandardItemModel()
                hm.appendRow(self.read_level(orient=role))
                return hm
            return super().data(index, role)

        def reorder(self, old_index, new_index, orientation) -> bool:
            """Reorder columns / rows."""
            horizontal = orientation == constants.HORIZONTAL
            cols = list(self.df.columns if horizontal else self.df.index)
            cols.insert(new_index, cols.pop(old_index))
            self.df = self.df[cols] if horizontal else self.df.T[cols].T
            return True

        # def headerData(self, section, orientation, role):
        #     if role != constants.DISPLAY_ROLE:
        #         return
        #     strings = (str(i) for i in idx)
        #     if orientation == constants.HORIZONTAL:
        #         idx = self.df.columns
        #         return "\n".join(strings) if isinstance(idx, tuple) else str(idx)
        #     else:
        #         idx = self.df.index
        #         return " | ".join(strings) if isinstance(idx, tuple) else str(idx)

    app = widgets.app()
    form = widgets.Widget()
    # form.setAttribute(Qt.WA_DeleteOnClose)  # http://stackoverflow.com/a/27178019/1119602
    form.setMinimumSize(700, 260)
    view = widgets.TableView()
    widgets.VBoxLayout(form).addWidget(view)
    form.show()

    # Prepare data
    tuples = [
        ("bar", "one", "q"),
        ("bar", "two", "q"),
        ("baz", "one", "q"),
        ("baz", "two", "q"),
        ("foo", "one", "q"),
        ("foo", "two", "q"),
        ("qux", "one", "q"),
        ("qux", "two", "q"),
    ]
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(6, 6), index=index[:6], columns=index[:6])

    # Prepare view
    #    oldh, oldv = view.horizontalHeader(), view.verticalHeader()
    #    oldh.setParent(form), oldv.setParent(form) #Save old headers for some reason
    HierarchicalHeaderView(constants.HORIZONTAL, view)
    HierarchicalHeaderView(constants.VERTICAL, view)
    view.h_header.setSectionsMovable(True)  # reorder DataFrame columns manually
    # Set data
    model = DataFrameModel(df)
    view.setModel(model)
    idx = model.index(0, 0)
    view.set_sorting_enabled(True, False)
    view.resizeColumnsToContents()
    view.resizeRowsToContents()
    # Set sorting enabled (after setting model)
    app.exec()



from __future__ import annotations

from prettyqt import widgets


class SizeGrip(widgets.WidgetMixin, widgets.QSizeGrip):
    pass


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Widget()
    widget = SizeGrip(w)
    widget.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import constants, core, custom_widgets, widgets


logger = logging.getLogger(__name__)

BOOL_ITEMS = {
    None: "Show all",
    constants.CheckState.Checked: "Show True",
    constants.CheckState.Unchecked: "Show False",
}


class FilterHeader(widgets.HeaderView):
    def __init__(self, parent: widgets.TableView):
        self._editors_visible = False
        self._editors = []
        self._proxy = parent.proxifier.get_proxy(
            "multi_column_filter",
            recursive_filtering_enabled=True,
        )
        self._padding = 6
        super().__init__(constants.HORIZONTAL, parent)
        self.setStretchLastSection(True)
        # self.setResizeMode(QHeaderView.Stretch)
        self.setDefaultAlignment(constants.ALIGN_CENTER_LEFT)
        # self.setSortIndicatorShown(False)
        self.sectionResized.connect(self._adjust_positions)
        parent.h_scrollbar.valueChanged.connect(self._adjust_positions)
        parent.model_changed.connect(self._update_filter_boxes)
        self.sectionResized.connect(self._adjust_positions)
        self._update_filter_boxes()
        self.update_geometries()

    def editors_visible(self) -> bool:
        return self._editors_visible

    def set_editors_visible(self, visible: bool):
        self._editors_visible = visible
        for editor in self._editors:
            editor.setVisible(visible)
        self.updateGeometries()

    def _update_filter_boxes(self):
        # TODO: deal with column changes by connecting to Model signals.
        # That way we wouldnt have to update all editors on change.
        while self._editors:
            editor = self._editors.pop()
            editor.deleteLater()
        self.create_editors()
        self._adjust_positions()

    def create_editors(self):
        # using parent model here bc we cant guarantee that we are already set to view.
        parent = self.parent()
        model = parent.model()
        self._proxy.clear_filters()
        for i in range(model.columnCount()):
            typ = model.get_column_type(i)
            if typ is bool:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val, constants.CHECKSTATE_ROLE)

                name = f"filter_combo_{i}"
                widget = widgets.ComboBox(margin=0, object_name=name, parent=self)
                widget.add_items(BOOL_ITEMS)
                widget.value_changed.connect(set_filter)
            elif typ in [int, float]:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val)

                name = f"filter_numwidget_{i}"
                widget = custom_widgets.NumFilterWidget(
                    margin=0, object_name=name, parent=self
                )
                widget.filter_changed.connect(set_filter)
                title = model.headerData(i, constants.HORIZONTAL, constants.DISPLAY_ROLE)
                widget.lineedit.setPlaceholderText(f"Filter {title}...")
            elif typ is str:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val)

                name = f"filter_lineedit_{i}"
                widget = widgets.LineEdit(margin=0, object_name=name, parent=self)
                widget.value_changed.connect(set_filter)
                title = model.headerData(i, constants.HORIZONTAL, constants.DISPLAY_ROLE)
                widget.setPlaceholderText(f"Filter {title}...")
            else:
                widget = widgets.Widget()
            widget.show()
            self._editors.append(widget)

    def sizeHint(self) -> core.QSize:
        size = super().sizeHint()
        if self._editors:
            height = self._editors[0].sizeHint().height()
            size.setHeight(size.height() + height + self._padding)
        return size

    def updateGeometries(self):
        if self._editors:
            height = self._editors[0].sizeHint().height()
            self.setViewportMargins(0, 0, 0, height + self._padding)
        else:
            self.setViewportMargins(0, 0, 0, 0)
        super().updateGeometries()
        self._adjust_positions()

    def _adjust_positions(self):
        for index, editor in enumerate(self._editors):
            height = editor.sizeHint().height()
            compensate_y = 0
            compensate_x = 0
            match editor:
                case widgets.QComboBox():
                    compensate_y = +2
                case widgets.QPushButton():
                    compensate_y = -1
                case widgets.QCheckBox():
                    compensate_y = 4
                    compensate_x = 4
                case widgets.QWidget():
                    compensate_y = -1
            editor.move(
                self.sectionPosition(index) - self.offset() + 1 + compensate_x,
                height + (self._padding // 2) + compensate_y,
            )
            editor.resize(self.sectionSize(index), height)

    def set_filter_case_sensitive(self, value: bool):
        self._proxy.set_filter_case_sensitive(value)

    def clear_filters(self):
        for editor in self._editors:
            editor.clear()

    editors_visible = core.Property(bool, editors_visible, set_editors_visible)


if __name__ == "__main__":
    from prettyqt import itemmodels

    app = widgets.app()
    with app.debug_mode():
        view = widgets.TableView()

        model = itemmodels.QObjectPropertiesModel(view, parent=view)
        # model = model.proxifier[:, 0:3]
        view.set_selection_behavior("rows")
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor", column=1)

        view.setModel(model)
        view.resize(640, 480)
        view.set_selection_behavior("rows")
        view.adapt_sizes()
        header = FilterHeader(parent=view)
        view.setHorizontalHeader(header)
        # view.h_header._update_filter_boxes()
        view.h_header.visible_editors = True
        view.show()
        with app.debug_mode():
            app.exec()
            print(view.h_header._proxy._filters)



from __future__ import annotations

from prettyqt.qt.QtScxml import *  # noqa: F403

from .scxmlcompiler import ScxmlCompiler
from scxmlstatemachine import ScxmlStateMachine
from .scxmldatamodel import ScXmlDataModel
from .scxmlcppdatamodel import ScXmlCppDataModel
from .scxmlnulldatamodel import ScXmlNullDataModel
from .scxmlinvokableservice import ScXmlInvokableService
from .scxmlinvokableservicefactory import ScXmlInvokableServiceFactory

__all__ = [
    "ScxmlCompiler",
    "ScxmlStateMachine",
    "ScXmlDataModel",
    "ScXmlCppDataModel",
    "ScXmlNullDataModel",
    "ScXmlInvokableService",
    "ScXmlInvokableServiceFactory",
]



"""pandaswidgets module.

contains Pandas-related widgets
"""

from .dataframelistwidget import DataFrameListWidget
from .dataframemanagerwidget import DataFrameManagerWidget
from .dataframeviewer import DataFrameViewer
from .dataframewidget import DataFrameWidget


__all__ = [
    "DataFrameViewer",
    "DataFrameWidget",
    "DataFrameListWidget",
    "DataFrameManagerWidget",
]



from __future__ import annotations

import logging

from typing import Any

import numpy as np
import pandas as pd

from prettyqt import constants, core, itemmodels
from prettyqt.utils import helpers


logger = logging.getLogger(__name__)


class PandasCategoryListModel(itemmodels.ModelMixin, core.AbstractTableModel):
    MIME_TYPE = "mime_categories"
    HEADER = ["Name", "Code"]

    def __init__(self, df: pd.DataFrame | None = None, col=None, parent=None):
        super().__init__(parent=parent)
        self.df = df
        self.col = col

    @property
    def series(self):
        return self.df[self.col]

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return False
        if role == constants.EDIT_ROLE and index.column() == 0:
            if value in self.series.cat.categories:
                logger.error("Error: Category names must be unique.")
                return False
            rename_dct = {index.data(constants.NAME_ROLE): value}
            with self.change_layout():
                self.series = self.series.rename_categories(new_categories=rename_dct)
            return True
        return False

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return constants.DROP_ENABLED
        return self.DEFAULT_FLAGS | constants.IS_EDITABLE

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        # if role == constants.DECORATION_ROLE:
        #     if index.column() == 0:
        #         return iconprovider.get_icon(iconnames.CATEGORY, as_qicon=True)
        cat_name = self.series.cat.categories[index.row()]
        match role, index.column():
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 0:
                return helpers.format_name(cat_name)
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 1:
                return str(self.series.cat.codes[index.row()])
            case constants.NAME_ROLE, _:
                if isinstance(cat_name, np.number):
                    return cat_name.item()
                return cat_name
            case constants.TOOLTIP_ROLE, _:
                num_items = len(self.df[self.series == cat_name])
                return f"<b>{cat_name}</b><br>Items: {num_items}"
            case constants.USER_ROLE, _:
                self.df = self.df[self.series == cat_name]
                return self.df

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        if not self.col or not hasattr(self.series, "cat"):
            return 0
        return len(self.series.cat.categories)

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        data = [i.row() for i in indexes if i.column() == 0]
        mime_data.set_json_data(self.MIME_TYPE, data)
        return mime_data

    def mimeTypes(self):
        return [self.MIME_TYPE]

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if parent_index.isValid():
            return False

        if mime_data.hasFormat(self.MIME_TYPE):
            indexes = mime_data.get_json_data(self.MIME_TYPE)
            cats = self.series.cat.categories.tolist()
            cats = helpers.move_in_list(cats, indexes, row)
            self.reorder_categories(cats)
            return False

    def sort(self, ncol, order):
        if ncol != 0:
            return None
        is_ascending = order == constants.ASCENDING
        order = sorted(self.series.cat.categories, reverse=is_ascending)
        self.reorder_categories(order)

    def reorder_categories(self, order):
        with self.change_layout():
            self.series = self.series.cat.reorder_categories(new_categories=order)

    def add_category(self, new_categories):
        with self.append_rows(1):
            self.series = self.series.cat.add_categories(new_categories=new_categories)

    def toggle_ordered(self):
        with self.change_layout():
            if self.series.cat.ordered:
                self.series = self.series.cat.set_unordered()
            else:
                self.series = self.series.cat.set_ordered()

    def remove_unused(self):
        with self.change_layout():
            self.series = self.series.cat.remove_unused()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    # Prepare data
    df = pd.DataFrame({"A": ["a", "b", "c", "a"]})

    df["B"] = df["A"].astype("category")
    model = PandasCategoryListModel(df, "B")
    tableview = widgets.TreeView()
    tableview.setup_dragdrop_move()
    # tableview.set_delegate("editor")
    tableview.set_model(model)

    tableview.show()
    app.exec()



from __future__ import annotations

import enum
import logging

from typing import Literal

from prettyqt import constants, core, gui, itemmodels, widgets
from prettyqt.utils import bidict


logger = logging.getLogger(__name__)


class SectionWidget(widgets.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_layout("vertical", margin=0, spacing=0)
        self.set_margin(0)
        header = widgets.Label(self.windowTitle()).set_bold()
        header.set_point_size(header.font().pointSize() * 2)
        self.box.add(header)
        for i in range(5):
            subitem = widgets.CheckBox(f"sub item {i}")
            self.box.add(subitem)


class ScrollAreaTocModel(itemmodels.TreeModel):
    def __init__(
        self, *args, widget_class: type[widgets.QWidget] = widgets.QWidget, **kwargs
    ):
        self._highlight_font = gui.QFont()
        self._highlight_font.setBold(True)
        self._Class = widget_class
        self._current_widgets: list[widgets.QWidget] = []
        self._header_property = "windowTitle"
        super().__init__(*args, **kwargs)

    def set_highlighted_widgets(self, widgets: list[widgets.QWidget]):
        self._current_widgets = widgets
        self.update_all()

    def columnCount(self, parent=None):
        return 1

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        widget = self.data_by_index(index).obj
        match role:
            case constants.DISPLAY_ROLE if widget in self._current_widgets:
                return f"> {widget.property(self.header_property)}"
            case constants.DISPLAY_ROLE:
                return widget.property(self.header_property)
            case constants.DECORATION_ROLE:
                return widget.windowIcon()
            case constants.USER_ROLE:
                return widget
            case constants.FONT_ROLE if widget in self._current_widgets:
                # if not index.parent().isValid():
                return self._highlight_font

    def _fetch_object_children(
        self, item: ScrollAreaTocModel.TreeItem
    ) -> list[ScrollAreaTocModel.TreeItem]:
        flag = constants.FindChildOption.FindDirectChildrenOnly
        children = item.obj.findChildren(self._Class, None, flag)
        children = [i for i in children if i.property(self.header_property)]
        return [self.TreeItem(obj=i) for i in children]

    def _has_children(self, item: ScrollAreaTocModel.TreeItem) -> bool:
        flag = constants.FindChildOption.FindDirectChildrenOnly
        children = item.obj.findChildren(self._Class, None, flag)
        children = [i for i in children if i.property(self.header_property)]
        return bool(children)

    def set_highlight_font(self, font: gui.QFont):
        self._highlight_font = font

    def get_highlight_font(self) -> gui.QFont:
        return self._highlight_font

    def set_header_property(self, prop: str):
        self._header_property = prop

    def get_header_property(self) -> str:
        return self._header_property

    highlight_font = core.Property(gui.QFont, get_highlight_font, set_highlight_font)
    header_property = core.Property(str, get_header_property, set_header_property)


class ScrollMode(enum.Enum):
    Single = 1
    Multi = 2
    HeadersOnly = 4


ScrollModeStr = Literal["single", "multi", "headers_only"]

SCROLL_MODE: bidict[ScrollModeStr, ScrollMode] = bidict(
    single=ScrollMode.Single, multi=ScrollMode.Multi, headers_only=ScrollMode.HeadersOnly
)


class ScrollAreaTocWidget(widgets.TreeView):
    section_changed = core.Signal()

    ScrollMode = ScrollMode
    core.Enum(ScrollMode)

    def __init__(
        self,
        scrollarea: widgets.QScrollArea,
        orientation: constants.Orientation
        | constants.OrientationStr = constants.VERTICAL,
        widget_class: type = widgets.QWidget,
        **kwargs,
    ) -> None:
        # TODO: not sure if parent should always equal scrollarea..."""
        self._WidgetClass = widget_class
        self._scroll_mode = ScrollMode.Single
        self._always_expanded = False
        self._last_visible = None
        self.scrollarea = scrollarea
        super().__init__(scrollarea, **kwargs)
        self._orientation = constants.ORIENTATION.get_enum_value(orientation)
        self.setFixedWidth(200)
        self.h_header.hide()
        self.h_header.setStretchLastSection(True)
        self.setAlternatingRowColors(False)
        self.setRootIsDecorated(False)
        # self.setStyleSheet(
        #     """::item:hover {background: transparent; border-color:transparent}
        #     ::item:selected { border-color:transparent;
        #     border-style:outset; border-width:2px; color:black; }"""
        # )
        if self._orientation == constants.VERTICAL:
            scrollarea.v_scrollbar.valueChanged.connect(self._on_scroll)
        else:
            scrollarea.h_scrollbar.valueChanged.connect(self._on_scroll)
        self.set_widget(scrollarea)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"scrollMode": SCROLL_MODE}
        return maps

    def showEvent(self, event):
        super().showEvent(event)
        self._on_scroll()

    def set_widget(self, widget: widgets.QScrollArea):
        """Set the ScrollArea widget to follow."""
        if widget.widget() is None:
            raise RuntimeError("No widget set on ScrollArea.")
        self.scrollarea = widget
        model = ScrollAreaTocModel(
            widget.widget(),
            show_root=True,
            parent=self.scrollarea,
            widget_class=self._WidgetClass,
        )
        self.set_model(model)
        self.proxy = self.proxifier.set_sort_filter_proxy(
            recursive_filtering_enabled=True
        )
        self.proxy.set_filter_case_sensitive(False)
        self.show_root(False)
        widget.widget().installEventFilter(self)
        self.selectionModel().currentRowChanged.connect(self._on_current_change)
        self.selectionModel().selectionChanged.connect(self._on_selection_change)
        # if self._always_expanded:
        self.expandAll()

    def _on_current_change(self, new, old):
        if self.model() is None:
            return
        is_vertical = self._orientation == constants.VERTICAL
        area = self.scrollarea
        scrollbar = area.v_scrollbar if is_vertical else area.h_scrollbar
        with self.signal_blocked(scrollbar.valueChanged, self._on_scroll):
            widget = self.model().data(new, role=constants.USER_ROLE)
            area.scroll_to_bottom()
            area.ensureWidgetVisible(widget, 10, 10)

    def _on_selection_change(self, new, old):
        if self.model() is None:
            return
        widgets = [i.data(constants.USER_ROLE) for i in self.selected_indexes()]
        self.get_model(skip_proxies=True).set_highlighted_widgets(widgets)

    def _on_scroll(self):
        model: ScrollAreaTocModel | None = self.model()
        if model is None:
            return
        visible_widgets = self.scrollarea.get_visible_widgets(typ=self._WidgetClass)
        if not visible_widgets or visible_widgets == self._last_visible:
            return
        self._last_visible = visible_widgets
        sig = self.selectionModel().currentRowChanged
        with self.signal_blocked(sig, self._on_current_change):
            self.select_index(None)
            if not self._always_expanded:
                self.collapseAll()
            match self.get_scroll_mode():
                case "multi":
                    indexes = model.search_tree(visible_widgets, constants.USER_ROLE)
                    for index in indexes:
                        children = model.get_child_indexes(index)
                        # only select if all children selected.
                        # if all(c in indexes for c in children):

                        # highlight when no children or when first child is visible.
                        if not children or children[0] in indexes:
                            self.select_index(index, clear=False)
                        self.set_expanded(indexes)
                        self.scroll_to(indexes[0])
                        self.scroll_to(indexes[-1])
                case "headers_only":
                    if indexes := model.search_tree(
                        visible_widgets,
                        role=constants.USER_ROLE,
                        max_results=1,
                    ):
                        self.set_current_index(indexes[0])
                        self.scroll_to(indexes[0])
                case "single":
                    if indexes := model.search_tree(visible_widgets, constants.USER_ROLE):
                        viewport = self.scrollarea.viewport()
                        # sort indexes by closest distance to top
                        indexes.sort(
                            key=lambda x: abs(
                                x.data(constants.USER_ROLE)
                                .map_to(viewport, x.data(constants.USER_ROLE).rect())
                                .top()
                            ),
                        )
                        self.collapseAll()
                        self.model().fetchMore(indexes[0])
                        self.set_current_index(indexes[0])
                        self.scroll_to(indexes[0])
                case _:
                    raise ValueError(self._scroll_mode)
        # model.set_highlighted_indexes(indexes)

    def wheelEvent(self, e):
        self.scrollarea.wheelEvent(e)

    def eventFilter(self, source: core.QObject, event: core.QEvent) -> bool:
        match event.type():
            case core.Event.Type.ChildAdded:
                self._on_scroll()
        return False

    def _scrollMode(self) -> ScrollMode:
        return self._scroll_mode

    def get_scroll_mode(self) -> ScrollModeStr:
        return SCROLL_MODE.inverse[self._scrollMode()]

    def set_scroll_mode(self, mode: ScrollMode | ScrollModeStr):
        self._scroll_mode = SCROLL_MODE.get_enum_value(mode)

    def is_always_expanded(self) -> bool:
        return self._always_expanded

    def set_always_expanded(self, always_expanded: bool):
        self._always_expanded = always_expanded

    scrollMode = core.Property(enum.Enum, _scrollMode, set_scroll_mode)
    always_expanded = core.Property(bool, is_always_expanded, set_always_expanded)


if __name__ == "__main__":
    app = widgets.app()
    with app.debug_mode():
        window = widgets.Widget(object_name="window")
        layout = window.set_layout("horizontal")
        scrollarea = widgets.ScrollArea(
            object_name="scroll", parent=window, widget_resizable=True
        )
        widget = widgets.Widget(object_name="scrollarea_widget")
        scrollarea.setWidget(widget)
        toc_widget = ScrollAreaTocWidget(
            scrollarea,
            scroll_mode="headers_only",
            widget_class=SectionWidget,
        )
        layout.add(toc_widget)
        scroll_layout = widget.set_layout("vertical")
        layout.add(scrollarea)
        for i in range(10):
            section = SectionWidget(window_title=f"test{i}")
            section.box.add(SectionWidget(window_title=f"{i}nested"))
            section.box.add(SectionWidget(window_title=f"{i*10}nested"))
            scroll_layout.add(section)
        window.show()
        app.exec()



from __future__ import annotations

import logging
import threading

import numpy as np
import pandas as pd

from prettyqt import constants, core, custom_widgets, gui, widgets
from prettyqt.qtpandas import pandasmodels


logger = logging.getLogger(__name__)

SelectionFlag = core.ItemSelectionModel.SelectionFlag


class DataTableView(widgets.TableView):
    def __init__(self, df=None, **kwargs):
        super().__init__(**kwargs)

        self.h_header.hide()
        self.v_header.hide()
        self.setAlternatingRowColors(True)
        self.set_scroll_mode("pixel")
        self.set_margin(0)
        self.set_delegate("no_focus")
        self.set_scrollbar_policy("always_off")
        self.set_delegate("editor")
        self.set_size_policy("expanding", "expanding")
        self.set_df(df)

    def set_df(self, df: pd.DataFrame):
        model = pandasmodels.DataTableModel(df, parent=self)
        self.set_model(model)

    def copy(self):
        indexes = self.selectionModel().selection().indexes()

        rows = [ix.row() for ix in indexes]
        cols = [ix.column() for ix in indexes]

        df = self.get_model(skip_proxies=True).df
        df = df.iloc[min(rows) : max(rows) + 1, min(cols) : max(cols) + 1]
        kwargs = dict(index=False, header=False)
        threading.Thread(target=df.to_clipboard, kwargs=kwargs).start()

    def paste(self):
        app = widgets.app()
        clipboard = app.clipboard()
        # TODO
        print(clipboard.text())

    def sizeHint(self):
        if self.get_model(skip_proxies=True).df is None:
            return core.Size(0, 0)
        width = 2 * self.frameWidth()  # Account for border & padding
        # width += self.v_scrollbar.width()
        max_cols = min(100, self.model().columnCount())
        for i in range(max_cols):
            width += self.columnWidth(i)

        # Height
        height = 2 * self.frameWidth()  # Account for border & padding
        # height += self.h_scrollbar.height()
        max_rows = min(100, self.model().rowCount())
        for i in range(max_rows):
            height += self.rowHeight(i)

        return core.Size(width, height)


class DataFrameViewer(widgets.Widget):
    def __init__(self, df: pd.DataFrame | None = None, **kwargs):
        super().__init__(**kwargs)
        # Indicates whether the widget has been shown yet. Set to True in
        self._loaded = False
        self.df = df
        # Set up DataFrame TableView and Model
        self.table_data = DataTableView(df, parent=self)
        self.set_size_policy("expanding", "expanding")

        self.table_data.selectionModel().selectionChanged.connect(
            self._on_selection_changed
        )
        # Create headers
        self.table_columns = HeaderView(orientation=constants.HORIZONTAL, parent=self)
        self.table_columns.clicked.connect(self.on_clicked)

        self.table_index = HeaderView(orientation=constants.VERTICAL, parent=self)

        # Set up layout
        self.layout_grid = self.set_layout("grid", margin=0, spacing=0)
        # Link scrollbars
        self.table_data.h_scrollbar.valueChanged.connect(
            self.table_columns.h_scrollbar.setValue
        )
        self.table_data.v_scrollbar.valueChanged.connect(
            self.table_index.v_scrollbar.setValue
        )
        self.table_columns.h_scrollbar.valueChanged.connect(
            self.table_data.h_scrollbar.setValue
        )
        self.table_index.v_scrollbar.valueChanged.connect(
            self.table_data.v_scrollbar.setValue
        )

        # Disable scrolling on the headers. Even when hidden, dragging desyncs them
        self.table_index.h_scrollbar.valueChanged.connect(lambda: None)

        # Add items to layout

        # SpacerItem |            Table_columns
        #       table index             | TrackingSpacer
        #       Table Index             | Table Data      | V Scrollbar
        #            |  Tracking Spacer  |H Scrollbar
        self.layout_grid[0, 0] = widgets.SpacerItem(0, 0, "expanding", "expanding")
        self.layout_grid[0, 1:2] = self.table_columns
        self.layout_grid[1:2, 0:1] = self.table_index
        self.layout_grid[1, 2] = TrackingSpacer(ref_y=self.table_index.h_header)
        self.layout_grid[2, 2] = self.table_data
        self.layout_grid[2, 3] = self.table_data.v_scrollbar
        self.layout_grid[3, 1] = TrackingSpacer(ref_x=self.table_columns.v_header)
        self.layout_grid[3, 2] = self.table_data.h_scrollbar

        # These expand when the window is enlarged instead of having
        # the grid squares spread out
        self.layout_grid.setColumnStretch(4, 1)
        self.layout_grid.setRowStretch(4, 1)

        # These placeholders will ensure the size of the blank spaces beside our headers
        if df is not None:
            self.set_df(df)

    def set_df(self, df: pd.DataFrame):
        self.df = df
        model = pandasmodels.DataTableModel(df, parent=self.table_data)
        self.table_data.set_model(model)

        self.table_data.selectionModel().selectionChanged.connect(
            self._on_selection_changed
        )
        # TODO: this gets called twice.
        self.table_columns.set_df(df)
        self.table_index.set_df(df)
        # Toggle level names
        if not (any(df.columns.names) or df.columns.name):
            self.table_columns.v_header.setFixedWidth(0)
        if not (any(df.index.names) or df.index.name):
            self.table_index.h_header.setFixedHeight(0)
        self.table_columns.selectionModel().selectionChanged.connect(
            self.on_column_selection_changed
        )
        self.table_index.selectionModel().selectionChanged.connect(
            self.on_index_selection_changed
        )
        self._resize_columns()
        self.table_data.updateGeometry()
        self.table_columns.updateGeometry()

    def jump_to_column(self, col_num: int):
        """Make sure column at given index is visible.

        Scrolls to column at given index.

        Arguments:
            col_num: column to scroll to
        """
        if (model := self.table_data.model()) is not None:
            idx = model.index(0, col_num)
            self.table_data.scrollTo(idx)
        if (model := self.table_index.model()) is not None:
            idx = model.index(0, col_num)
            self.table_index.scrollTo(idx)
        if (model := self.table_columns.model()) is not None:
            idx = model.index(0, col_num)
            self.table_columns.scrollTo(idx)

    # def on_column_selection_changed(self, selected, deselected):
    #     if self.table_columns.hasFocus():
    #         selected = self.table_columns.get_higher_levels()
    #         self.table_data.selectionModel().select(
    #             selected,
    #             SelectionFlag.Columns
    #             | SelectionFlag.ClearAndSelect,
    #         )
    #     self.table_columns._on_selection_changed(selected, deselected)

    # def on_index_selection_changed(self, selected, deselected):
    #     if not self.table_index.hasFocus():
    #         return
    #     selected = self.table_index.get_higher_levels()
    #     self.table_data.selectionModel().select(
    #         selected,
    #         SelectionFlag.Rows
    #         | SelectionFlag.ClearAndSelect,
    #     )
    #     self.table_index._on_selection_changed(selected, deselected)

    def on_clicked(self, ix: core.QModelIndex):
        self.df = self.df.sort_values(self.df.columns[ix.column()])
        self.data_changed()

    def showEvent(self, event: gui.QShowEvent):
        """Initialize column and row sizes on the first time the widget is shown."""
        if not self._loaded and self.df is not None:
            # Set column widths
            self._resize_columns()

            self._loaded = True
        event.accept()

    def _resize_columns(self):
        N = 50
        columns = min(self.table_index.model().columnCount(), N)
        for column_index in range(columns):
            self.auto_size_column(column_index)

        # Set row heights
        # Just sets a single uniform row height based on the first
        # N rows for performance.
        default_row_height = 30
        rows = min(self.table_index.model().rowCount(), N)
        for row_index in range(rows):
            self.auto_size_row(row_index)
            height = self.table_index.rowHeight(row_index)
            default_row_height = max(default_row_height, height)

        # Set limit for default row height
        default_row_height = min(default_row_height, 100)

        self.table_index.v_header.setDefaultSectionSize(default_row_height)
        self.table_data.v_header.setDefaultSectionSize(default_row_height)

    def on_column_selection_changed(self, selected, deselected):
        if not self.table_columns.hasFocus():
            return
        higher_levels = self.table_columns.get_higher_levels()
        # Set selection mode so selecting one row or column at a time
        # adds to selection each time
        selected.merge(higher_levels, SelectionFlag.Deselect)
        # Select the cells in the data view
        self.table_data.selectionModel().select(
            selected, SelectionFlag.Columns | SelectionFlag.ClearAndSelect
        )

    def on_index_selection_changed(self, selected, deselected):
        if not self.table_index.hasFocus():
            return
        higher_levels = self.table_index.get_higher_levels()
        selected.merge(higher_levels, SelectionFlag.Deselect)
        self.table_data.selectionModel().select(
            selected, SelectionFlag.Rows | SelectionFlag.ClearAndSelect
        )

    def _on_selection_changed(self, selected, deselected):
        """Runs when cells are selected in the main table.

        This logic highlights the correct
        cells in the vertical and horizontal headers when a data cell is selected
        """
        # The two blocks below check what columns or rows are selected in the data table
        # and highlights thecorresponding ones in the two headers. The if statements
        # check for focus on headers, because if the user clicks a header that will
        # auto-select all cells in that row or column which will trigger this function
        # and cause and infinite loop

        if not self.table_columns.hasFocus():
            self.table_columns.selectionModel().select(
                selected, SelectionFlag.Columns | SelectionFlag.ClearAndSelect
            )

        if not self.table_index.hasFocus():
            self.table_index.selectionModel().select(
                selected, SelectionFlag.Rows | SelectionFlag.ClearAndSelect
            )

    def auto_size_column(self, column_index: int):
        """Set the size of column at column_index to fit its contents."""
        padding = 20

        self.table_columns.resizeColumnToContents(column_index)
        width = self.table_columns.columnWidth(column_index)

        # Iterate over the column's rows and check the width of each
        # to determine the max width for the column
        # Only check the first N rows for performance. If there is larger
        # content in cells below it will be cut off
        N = 75
        rows = min(self.table_data.model().rowCount(), N)
        for i in range(rows):
            mi = self.table_data.model().index(i, column_index)
            text = self.table_data.model().data(mi)
            text = str(text)  # TODO: get rid of casting?
            w = self.table_data.fontMetrics().boundingRect(text).width()

            width = max(width, w)

        width += padding

        # add maximum allowable column width so column is never too big.
        max_allowable_width = 400
        width = min(width, max_allowable_width)

        self.table_columns.setColumnWidth(column_index, width)
        self.table_data.setColumnWidth(column_index, width)

        self.table_data.updateGeometry()
        self.table_columns.updateGeometry()

    def auto_size_row(self, row_index: int):
        """Set the size of row at row_index to fix its contents."""
        padding = 20

        self.table_index.resizeRowToContents(row_index)
        height = self.table_index.rowHeight(row_index)

        # Iterate over the row's columns and check the width of each to
        # determine the max height for the row
        # Only check the first N columns for performance.
        N = 100
        model = self.table_data.model()
        for i in range(min(N, model.columnCount())):
            mi = model.index(row_index, i)
            cell_width = self.table_columns.columnWidth(i)
            text = model.data(mi)
            # Gets row height at a constrained width (the column width).
            # This constrained width, with the flag of Qt.TextWordWrap
            # gets the height the cell would have to be to fit the text.
            constrained_rect = core.Rect(0, 0, cell_width, 0)
            text = str(text)  # TODO: remove this?
            h = (
                self.table_data.fontMetrics()
                .boundingRect(constrained_rect, constants.TextFlag.TextWordWrap, text)
                .height()
            )

            height = max(height, h)

        height += padding

        self.table_index.setRowHeight(row_index, height)
        self.table_data.setRowHeight(row_index, height)

        self.table_data.updateGeometry()
        self.table_index.updateGeometry()

    def keyPressEvent(self, event: core.QEvent):
        super().keyPressEvent(event)

        if event.matches(gui.KeySequence.StandardKey.Copy):
            self.table_data.copy()
        elif event.matches(gui.KeySequence.StandardKey.Paste):
            self.table_data.paste()

    def data_changed(self):
        # Call dataChanged on all models for all data
        for model in [
            self.table_data.model(),
            self.table_columns.model(),
            self.table_index.model(),
        ]:
            model.update_all()


class HeaderView(custom_widgets.OrientedTableView):
    """Displays the DataFrame index or columns depending on orientation."""

    def __init__(self, parent: DataFrameViewer, orientation: constants.Orientation):
        super().__init__(orientation, parent=parent)

        self.table = parent.table_data
        # These are used during column resizing
        self._header_is_resizing = None
        self.resize_start_pos: int | None = None
        self.initial_header_size = None

        # Handled by self.eventFilter()
        self.setMouseTracking(True)
        self.viewport().setMouseTracking(True)
        self.viewport().installEventFilter(self)

        # Settings
        self.set_size_policy("maximum", "maximum")
        self.setWordWrap(False)
        self.set_horizontal_scrollbar_policy("always_off")
        self.set_scroll_mode("pixel")
        self.set_margin(0)
        self.set_delegate("no_focus")

        # Orientation specific settings
        if self.is_horizontal():
            self.set_horizontal_scrollbar_policy("always_off")
            self.h_header.hide()
            self.v_header.setDisabled(True)
            # Selection lags without this
            self.v_header.setHighlightSections(False)

        else:
            self.set_vertical_scrollbar_policy("always_off")
            self.v_header.hide()
            self.h_header.setDisabled(True)
            self.h_header.setHighlightSections(False)
        df = parent.df
        if df is not None:
            self.set_df(df)
            self.resize(self.sizeHint())

    def set_df(self, df: pd.DataFrame, set_spans: bool = True):
        if self.is_horizontal():
            model = pandasmodels.HorizontalHeaderModel(df)
        else:
            model = pandasmodels.VerticalHeaderModel(df)
        self.set_model(model)
        # Link selection to DataTable
        if set_spans:
            self.set_spans()
        self.init_size()
        # Set initial size
        self.resize(self.sizeHint())

    def get_higher_levels(self):
        model = self.model()
        if self.is_horizontal():
            # Get the header's selected columns
            # Removes the higher levels so that only the lowest level of the header
            # affects the data table selection
            last_row_ix = model.df.columns.nlevels - 1
            last_col_ix = model.columnCount() - 1
            return core.ItemSelection(
                model.index(0, 0), model.index(last_row_ix - 1, last_col_ix)
            )
        else:
            last_row_ix = model.rowCount() - 1
            last_col_ix = model.df.index.nlevels - 1
            return core.ItemSelection(
                model.index(0, 0), model.index(last_row_ix, last_col_ix - 1)
            )

    # Take the current set of selected cells and make it so that any spanning cell above
    # a selected cell is selected too This should happen after every selection change
    def selectionChanged(self, selected, deselected):
        if self.is_horizontal():
            if self.get_model(skip_proxies=True).df.columns.nlevels == 1:
                return
            for ix in self.selectedIndexes():
                for row in range(ix.row()):
                    ix2 = self.model().index(row, ix.column())
                    self.setSelection(self.visualRect(ix2), SelectionFlag.Select)
        else:
            if self.get_model(skip_proxies=True).df.index.nlevels == 1:
                return
            for ix in self.selectedIndexes():
                for col in range(ix.column()):
                    ix2 = self.model().index(ix.row(), col)
                    self.setSelection(self.visualRect(ix2), SelectionFlag.Select)

    # Fits columns to contents but with a minimum width and added padding
    def init_size(self):
        padding = 20
        self.resizeColumnsToContents()
        colcount = min(self.model().columnCount(), 100)
        if self.is_horizontal():
            min_size = 100
            for col in range(colcount):
                width = self.columnWidth(col)
                new_width = max(width + padding, min_size)
                self.setColumnWidth(col, new_width)
                self.table.setColumnWidth(col, new_width)
        else:
            # max_size = 1000
            for col in range(colcount):
                width = self.columnWidth(col)
                self.setColumnWidth(col, width + padding)

    # This sets spans to group together adjacent cells with the same values
    def set_spans(self):
        df = self.get_model(skip_proxies=True).df
        index = df.columns if self.is_horizontal() else df.index
        is_multiindex = isinstance(index, pd.MultiIndex)
        n = len(index[0]) if is_multiindex else 1
        for level in range(n):
            # Find how many segments the MultiIndex has
            arr = index.codes[level] if is_multiindex else index
            starts = np.where(arr[:-1] != arr[1:])[0] + 1
            starts = np.insert(starts, 0, 0)
            lengths = np.diff(starts)
            lengths = np.append(lengths, len(arr) - starts[-1])
            for start, length in zip(starts, lengths):
                if length > 1:
                    self.set_section_span(level, start, length)
            # Holds the starting index of a range of equal values.
            # None means it is not currently in a range of equal values.
            # match_start = None
            # for col in range(1, len(arr)):  # Iterates over cells in row
            #     # Check if cell matches cell to its left
            #     if arr[col] == arr[col - 1]:
            #         if match_start is None:
            #             match_start = col - 1
            #         # If this is the last cell, need to end it
            #         if col == len(arr) - 1:
            #             match_end = col
            #             span_size = match_end - match_start + 1
            #             self.set_section_span(level, match_start, span_size)
            #     elif match_start is not None:
            #         match_end = col - 1
            #         span_size = match_end - match_start + 1
            #         self.set_section_span(level, match_start, span_size)
            #         match_start = None

    def eventFilter(self, source: core.QObject, event: core.QEvent):
        if not isinstance(event, gui.QMouseEvent):
            return False
        mouse_pos = event.position().x() if self.is_horizontal() else event.position().y()
        mouse_pos = int(mouse_pos)
        # If mouse is on an edge, start the drag resize process
        match event.type():
            case core.QEvent.Type.MouseButtonPress:
                if self.over_header_edge(mouse_pos) is not None:
                    self._header_is_resizing = self.over_header_edge(mouse_pos)
                    self.resize_start_pos = mouse_pos
                    self.initial_header_size = self.sectionWidth(self._header_is_resizing)
                    return True
                else:
                    self._header_is_resizing = None
            # End the drag process
            case core.QEvent.Type.MouseButtonRelease:
                self._header_is_resizing = None

            case core.QEvent.Type.MouseButtonDblClick if self.is_horizontal():
                if (header_index := self.over_header_edge(mouse_pos)) is not None:
                    self.parent().auto_size_column(header_index)
                    return True

            case core.QEvent.Type.MouseButtonDblClick:
                if (header_index := self.over_header_edge(mouse_pos)) is not None:
                    self.parent().auto_size_row(header_index)
                    return True
            case core.QEvent.Type.MouseMove:
                # If this is None, there is no drag resize happening
                if self._header_is_resizing is not None:
                    size = self.initial_header_size + mouse_pos - self.resize_start_pos
                    if size > 10:
                        table_data = self.parent().table_data
                        self.setSectionWidth(self._header_is_resizing, size)
                        if self.is_horizontal():
                            table_data.setColumnWidth(self._header_is_resizing, size)
                        else:
                            table_data.setRowHeight(self._header_is_resizing, size)
                        self.updateGeometry()
                        table_data.updateGeometry()
                    return True

                # Set the cursor shape
                if self.over_header_edge(mouse_pos) is not None:
                    self.viewport().setCursor(self.get_split_cursor())
                else:
                    cursor = gui.Cursor(constants.CursorShape.ArrowCursor)
                    self.viewport().setCursor(cursor)
        return False

    # Return the size of the header needed to match the corresponding DataTableView
    def sizeHint(self):
        if self.model() is None:
            return core.Size(0, 0)
        # Horizontal HeaderView
        if self.is_horizontal():
            # Width of DataTableView
            width = self.table.sizeHint().width() + self.v_header.width()
            # Height
            height = 2 * self.frameWidth()  # Account for border & padding
            max_rows = min(100, self.model().rowCount())
            for i in range(max_rows):
                height += self.rowHeight(i)

        # Vertical HeaderView
        else:
            # Height of DataTableView
            height = self.table.sizeHint().height() + self.h_header.height()
            # Width
            width = 2 * self.frameWidth()  # Account for border & padding
            max_cols = min(100, self.model().columnCount())
            for i in range(max_cols):
                width += self.columnWidth(i)
        return core.Size(width, height)

    # This is needed because otherwise when the horizontal header is a single row
    # it will add whitespace to be bigger
    def minimumSizeHint(self):
        if self.is_horizontal():
            return core.Size(0, self.sizeHint().height())
        else:
            return core.Size(self.sizeHint().width(), 0)


# This is a fixed size widget with a size that tracks some other widget
class TrackingSpacer(widgets.Frame):
    def __init__(self, ref_x=None, ref_y=None):
        super().__init__()
        self.ref_x = ref_x
        self.ref_y = ref_y

    def minimumSizeHint(self):
        width = self.ref_x.width() if self.ref_x else 0
        height = self.ref_y.height() if self.ref_y else 0
        return core.Size(width, height)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    df = debugging.example_multiindex_df()
    with app.debug_mode():
        view2 = DataFrameViewer(df)
        view2.show()
        app.exec()



from __future__ import annotations

from prettyqt.qt.QtMultimediaWidgets import *  # noqa: F403

from .graphicsvideoitem import GraphicsVideoItem
from .videowidget import VideoWidget


__all__ = ["VideoWidget", "GraphicsVideoItem"]



"""Module containing custom animation classes."""

from .cursormoveanimation import CursorMoveAnimation
from .slideanimation import SlideAnimation
from .textanimation import TextAnimation
from .zoomanimation import ZoomAnimation


__all__ = ["SlideAnimation", "ZoomAnimation", "TextAnimation", "CursorMoveAnimation"]



from __future__ import annotations

from prettyqt.qt.QtPrintSupport import *  # noqa: F403

from .printer import Printer
from .abstractprintdialog import AbstractPrintDialog, AbstractPrintDialogMixin
from .printdialog import PrintDialog
from .pagesetupdialog import PageSetupDialog
from .printengine import PrintEngine
from .printpreviewwidget import PrintPreviewWidget
from .printpreviewdialog import PrintPreviewDialog

__all__ = [
    "Printer",
    "AbstractPrintDialog",
    "AbstractPrintDialogMixin",
    "PrintDialog",
    "PageSetupDialog",
    "PrintEngine",
    "PrintPreviewWidget",
    "PrintPreviewDialog",
]



from __future__ import annotations

import enum
import logging

from typing import Any

import numpy as np
import pandas as pd

from prettyqt import constants, core, itemmodels
from prettyqt.utils import helpers


logger = logging.getLogger(__name__)


class PandasColumnListModel(itemmodels.ModelMixin, core.AbstractTableModel):
    HEADER = ["Name", "Type", "Size"]
    MIME_TYPE = "mime_columns"
    content_type = "features"

    class Roles(enum.IntEnum):
        """Role names."""

        DTypeRole = constants.USER_ROLE + 4555
        ColumnNameRole = constants.USER_ROLE + 4556

    def __init__(self, df: pd.DataFrame = None, **kwargs):
        super().__init__(**kwargs)
        self.df = df

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        data = [i.row() for i in indexes if i.column() == 0]
        mime_data.set_json_data(self.MIME_TYPE, data)
        return mime_data

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if parent_index.isValid():
            # Since we only drop in between items, parent_index must be invalid
            return False
        if mime_data.hasFormat(self.MIME_TYPE):
            data = mime_data.get_json_data(self.MIME_TYPE)
            cols = self.df.columns.tolist()
            cols = helpers.move_in_list(cols, data, row)
            with self.change_layout():
                self.df = self.df[cols]
            self.update_all()
            return False

        elif mime_data.hasFormat("mime_indexes"):
            data = mime_data.get_json_data("mime_indexes")
            names = list(self.df.index.names)
            with self.reset_model():
                self.df = self.df.reset_index(level=[names[i] for i in data])
            self.update_all()
            return False

    def mimeTypes(self):
        return [self.MIME_TYPE]

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return False
        match role:
            case constants.EDIT_ROLE:
                colname = index.data(constants.NAME_ROLE)
                if value not in self.df.columns:
                    self.df = self.df.rename(mapper={colname: value}, axis="columns")
                    self.update_all()
                    return True

            case constants.USER_ROLE:
                name = self.data_by_index(index).name
                self.df = self.df.assign(**{name: value})
                self.update_all()
                return True

            case self.Roles.DTypeRole:
                name = index.data(constants.NAME_ROLE)
                try:
                    self.df = self.df.as_type(dtype=value)
                except TypeError as e:
                    logger.exception(e)
                else:
                    self.update_all()
                return True

        return False

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return constants.DROP_ENABLED
        if index.column() in [0, 1]:
            return self.DEFAULT_FLAGS | constants.IS_EDITABLE
        return self.DEFAULT_FLAGS

    def removeRows(self, row, count, parent):
        colnames = self.df.columns[row : row + count].tolist()
        with self.remove_rows(row, row + count, parent):
            self.df = self.df.drop(labels=colnames, axis=1, errors="ignore")
        self.update_all()
        return True

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None

        match role:
            # case constants.DECORATION_ROLE:
            #     if index.column() == 0:
            #         dtype = self.data_by_index(index).dtype
            #         return icons.icon_for_dtype(dtype)

            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                match index.column():
                    case 0:
                        return str(self.data_by_index(index).name)
                    case 1:
                        return str(self.data_by_index(index).dtype)
                    case 2:
                        nbytes = self.data_by_index(index).nbytes
                        locale = core.Locale()
                        return locale.get_formatted_data_size(nbytes)

            case self.Roles.DTypeRole:
                return self.data_by_index(index).dtype

            case constants.USER_ROLE:
                return self.data_by_index(index)

            case self.Roles.ColumnNameRole:
                return self.data_by_index(index).name

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        return len(self.df.columns) if self.df is not None else 0

    def sort(self, ncol, order):
        is_ascending = order == constants.ASCENDING
        match ncol:
            case 0:
                order = sorted(self.df.columns, reverse=is_ascending)

            case 1:
                maps = [(col, str(self.df[col].dtype)) for col in self.df.columns]
                maps.sort(key=lambda x: x[1], reverse=is_ascending)
                order = [c[0] for c in maps]

            case 2:
                maps = [(col, str(self.df[col].nbytes)) for col in self.df.columns]
                maps.sort(key=lambda x: x[1], reverse=is_ascending)
                order = [c[0] for c in maps]
            case _:
                return

        with self.change_layout():
            self.df = self.df.reindex(labels=order, axis="columns")

    def data_by_index(self, index: core.ModelIndex):
        return self.df.iloc[:, index.row()]

    def add_feature(self, series):
        with self.append_rows(1):
            self.df = self.df.assign(**{series.name: series})


def label_for_index(index: pd.Index) -> str:
    match index:
        case pd.RangeIndex():
            return "Range"
        case pd.DatetimeIndex():
            return "Datetime"
        case pd.CategoricalIndex():
            return "Categorical"
        case pd.IntervalIndex():
            return "Interval"
        case pd.TimedeltaIndex():
            return "Timedelta"
        case pd.MultiIndex():
            return "Multi-index"
        case pd.PeriodIndex():
            return "Period"
        case _:
            return "Regular"


class PandasIndexListModel(itemmodels.ModelMixin, core.AbstractTableModel):
    HEADER = ["Name", "Index type", "Monotonic"]
    MIME_TYPE = "mime_indexes"

    def __init__(self, ds: pd.DataFrame | None = None, **kwargs):
        super().__init__(**kwargs)
        self.df = ds

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return False
        match role:
            case constants.EDIT_ROLE:
                if index.column() == 0:
                    level = (
                        index.row() if isinstance(self.df.index, pd.MultiIndex) else None
                    )
                    self.df.index = self.df.index.set_names(names=value, level=level)
                    return True
            case constants.USER_ROLE:
                if not isinstance(self.df.index, pd.MultiIndex):
                    with self.change_layout():
                        self.df.index = value
                    return True
            case self.DTYPE_ROLE:
                try:
                    with self.change_layout():
                        self.df.index = self.df.index.astype(value)
                    return True
                except TypeError as e:
                    logger.exception(e)
                    return False

    def removeRows(self, row, count, parent):
        with self.remove_rows(row, row + count, parent):
            self.df = self.df.drop(labels=self.df.columns[row : row + count], axis=1)
        return True

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return constants.DROP_ENABLED
        is_range = self.df.index.name is None and isinstance(self.df.index, pd.RangeIndex)
        match index.column(), is_range:
            case 0, True:
                # do not allow moving the "default" RangeIndex.
                return (
                    constants.IS_ENABLED | constants.IS_SELECTABLE | constants.IS_EDITABLE
                )
            case _, True:
                return constants.IS_ENABLED | constants.IS_SELECTABLE
            case 0 | 1, False:
                return self.DEFAULT_FLAGS | constants.IS_EDITABLE
            case _, False:
                return self.DEFAULT_FLAGS

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        match role, index.column():
            # case constants.DECORATION_ROLE, 0:
            #     index = self.get_index(index.row())
            #     return icons.icon_for_index(index)

            case constants.TOOLTIP_ROLE, _:
                index = self.get_index(index.row())
                return (
                    f"<b>{index.name}</b><br>"
                    f"{'Contains NaNs'}: {index.hasnans}<br>"
                    # f"{_('Monotonic')}: {index.is_monotonic}"
                )

            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 0:
                index_name = self.df.index.names[index.row()]
                return index_name or "[Unnamed]"

            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 1:
                return label_for_index(self.get_index(index.row()))

            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 2:
                index = self.get_index(index.row())
                if index.is_monotonic_increasing:
                    return "Increasing"
                elif index.is_monotonic_decreasing:
                    return "Decreasing"

            case constants.USER_ROLE, _:
                name = self.df.index.names[index.row()]
                return self.df.index.get_level_values(name)

            case constants.NAME_ROLE, _:
                return self.df.index.names[index.row()]

            case self.DTYPE_ROLE, _:
                return self.get_index(index.row()).dtype

    def rowCount(self, parent=None):
        return self.df.index.nlevels if self.df is not None else 0

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        data = [i.row() for i in indexes if i.column() == 0]
        mime_data.set_json_data(self.MIME_TYPE, data)
        return mime_data

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if parent_index.isValid():
            return False

        if mime_data.hasFormat(self.MIME_TYPE):
            data = mime_data.get_json_data(self.MIME_TYPE)
            names = list(self.df.index.names)
            names = helpers.move_in_list(names, data, row)
            with self.change_layout():
                self.df = self.df.reorder_levels(order=names)
        elif mime_data.hasFormat("mime_columns"):
            data = mime_data.get_json_data("mime_columns")
            cols = self.df.columns.tolist()
            to_append = [cols[i] for i in data]
            # TODO: make sure drop offset is taken into account
            with self.reset_model():
                self.df = self.df.set_index(keys=to_append, append=True)
        return False

    def mimeTypes(self):
        return [self.MIME_TYPE]

    def get_index(self, row=None):
        idx = self.df.index
        # using get_level_values() is too slow for large datasets with multiindex
        return idx.levels[row] if isinstance(idx, pd.MultiIndex) else idx

    def to_datetime(self, level=None, fmt: str | None = None):
        with self.change_layout():
            self.df.index = self.df.index.pt.to_datetime(level=level, fmt=fmt)

    def filter_range(self, start: int, end: int):
        with self.change_layout():
            self.df = self.df.iloc[start:end]
        return True


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    # Prepare data
    tuples = [
        ("bar", "one", "q"),
        ("bar", "two", "q"),
        ("baz", "one", "q"),
        ("baz", "two", "q"),
        ("foo", "one", "q"),
        ("foo", "two", "q"),
        ("qux", "one", "q"),
        ("qux", "two", "q"),
    ]
    index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
    df = pd.DataFrame(np.random.randn(6, 6), index=index[:6], columns=index[:6])
    model = PandasColumnListModel(df)
    tableview = widgets.TreeView()
    tableview.setup_dragdrop_move()
    tableview.set_delegate("editor")
    tableview.set_model(model)

    tableview.show()
    app.exec()



"""svg module.

contains QtSvg-based classes
"""

from prettyqt.qt.QtSvg import *  # noqa: F403

from .svggenerator import SvgGenerator
from .svgrenderer import SvgRenderer


__all__ = ["SvgGenerator", "SvgRenderer"]
